"""
A collection of NMR spectral processing functions which operate on the last
dimension (1) of 2D arrays.  These functions are wrapped by other processing
modules but can also be used directly.  All parameter are assumed to be in
units of points unless otherwise noted.
"""

# TODO determine which of these work on N-dimension and which assume 2D

import numpy as np
import scipy.signal
import scipy.linalg

pi = np.pi


#########################
# Apodization functions #
#########################


def em(data, lb=0.0, inv=False, rev=False):
    """
    Exponential apodization

    Functional form of apodization window:

    .. math::
        em(x_i) = \\exp(-pi * i * lb)


    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    lb : float
        Exponential line broadening, in units of points. To apply a similar
        apodization as NMRPipe's EM function, use lb = lb_hz / sw_hz,
        where lb_hz is the amount of broadening to apply in Hz and sw_hz
        is the spectral width of the last dimension in Hz.
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    apod = np.exp(-pi * np.arange(data.shape[-1]) * lb).astype(data.dtype)
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod   # invert apodization
    return apod * data


def gm(data, g1=0.0, g2=0.0, g3=0.0, inv=False, rev=False):
    """
    Lorentz-to-Gauss apodization

    Functional form of apodization window:

    .. math::
        gm(x_i) = \\exp(e - g^2)

    Where:

    .. math::
        e = pi * i * g1 \\\\
        g = 0.6 * pi * g2 * (g3 * (size - 1) - i)

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    g1 : float
        Inverse exponential width.
    g2 : float
        Gaussian broadening width.
    g3 : float
        Location of Gaussian maximum.
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    size = data.shape[-1]
    e = pi * np.arange(size) * g1
    g = 0.6 * pi * g2 * (g3 * (size - 1) - np.arange(size))
    apod = np.exp(e - g * g).astype(data.dtype)
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod
    return apod * data


def gmb(data, a=0.0, b=0.0, inv=False, rev=False):
    """
    Modified gaussian apodization

    Functional form of apodization window:

    .. math::
        gmb(x_i) = \\exp(-a * i - b * i^2)

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    a : float
        Exponential term in apodization.
    b : float
        Gaussian term in apodization.
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    size = data.shape[-1]
    apod = np.exp(-a * np.arange(size) - b *
                  np.arange(size) ** 2).astype(data.dtype)
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod
    return apod * data


def jmod(data, e=0.0, off=0.0, end=0.0, inv=False, rev=False):
    """
    Exponentially damped J-modulation apodization

    Functional form of apodization window:

    .. math::
        jmod(x_i) = \\exp(-e) * \\sin(\\frac{pi * off + pi * (end - off) * i}
                                            {size - 1})

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    e : float
        Exponential apodization term.
    off : float
        Start of J-modulation in fractions of pi radians (180 degrees).
    end : float
        End of J-modulation in fractions of pi radians (180 degrees).
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    size = data.shape[-1]
    apod = (np.exp(-e * np.arange(size)).astype(data.dtype) *
            np.sin(pi * off + pi * (end - off) * np.arange(size) /
                   (size - 1)).astype(data.dtype))
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod
    return apod * data


def sp(data, off=0, end=1.0, pow=1.0, inv=False, rev=False):
    """
    Shifted sine-bell apodization

    Functional form of apodization window:

    .. math::
        sp(x_i) = \\sin(\\frac{pi * off + pi * (end - off) * i}
                              {size - 1})^{pow}

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    off : float
        Offset start of sine-bell as a percentage of the trace (0 -> 1).
    end : float
        End of of sine-bell as a percentage of the trace (0 -> 1).
    pow : float
        Power to raise sine-bell to.
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    size = data.shape[-1]
    apod = np.power(np.sin(pi * off + pi * (end - off) * np.arange(size) /
                    (size - 1)).astype(data.dtype), pow).astype(data.dtype)
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod
    return apod * data


sine = sp


def tm(data, t1=0.0, t2=0.0, inv=False, rev=False):
    """
    Trapezoid Apodization

    Functional form of apodization:

    ============ ================================
    Range        Description
    ============ ================================
    0:t1         Linear increases from 0.0 to 1.0
    t1:size - t2 Flat with value of 1.0
    -t2:         Linear descrease from 1.0 to 0.0
    ============ ================================

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    t1 : int
        Length of left ramp in points.
    t2 : int
        Length of right ramp in points.
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    size = data.shape[-1]
    apod = np.concatenate((np.linspace(0, 1, t1), np.ones(size - t1 - t2),
                           np.linspace(1, 0, t2))).astype(data.dtype)
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod
    return apod * data


def tri(data, loc="auto", lHi=0.0, rHi=0.0, inv=False, rev=False):
    """
    Triangle apodization.

    Functional form of apodization window.

    ============ ========================================
    Range        Description
    ============ ========================================
    0:loc        Linear increase/decrease from lHi to 1.0
    loc:         Linear increase/decrease from 1.0 to rHi
    ============ ========================================

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    loc : int or "auto"
        Location of triangle apex. 'auto' set the apex to the middle of the
        trace (default).
    lHi : float
        Starting height of the left side of the triangle.
    rHi : float
        Starting height of the right side of the triangle.
    inv : bool, optional
        True for inverse apodization.  False (default) for standard.
    rev : bool, optional.
        True to reverse the apodization before applying it to the data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with apodization applied.

    """
    size = data.shape[-1]
    if loc == "auto":
        loc = int(size / 2.)
    apod = np.concatenate((np.linspace(lHi, 1., loc), np.linspace(1., rHi,
                          size - loc + 1)[1:])).astype(data.dtype)
    if rev:
        apod = apod[::-1]
    if inv:
        apod = 1 / apod
    return apod * data


###################
# Shift functions #
###################


def rs(data, pts=0.0):
    """
    Right shift and zero fill.

    Parameters
    ----------
    data :
        Array of NMR data.
    pts : int
        Number of points to right shift.

    Returns
    -------
    ndata : ndarray
        Array of NMR data right shifted and zero filled.

    See Also
    --------
    roll : shift without zero filling.

    """
    data = np.roll(data, int(pts), axis=-1)
    data[..., :int(pts)] = 0
    return data


def ls(data, pts=0.0):
    """
    Left shift and fill with zero

    Parameters
    ----------
    data :
        Array of NMR data.
    pts : int
        Number of points to left shift.

    Returns
    -------
    ndata : ndarray
        Array of NMR data left shifted and zero filled.

    See Also
    --------
    roll : shift without zero filling.

    """
    data = np.roll(data, -int(pts), axis=-1)
    data[..., -int(pts):] = 0
    return data


def cs(data, pts=0.0, neg=False):
    """
    Circular shift

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    pts : int
        Number of points to shift.  Positive value will right shift the data,
        negative values will left shift the data.
    neg : bool
        True to negate the shifted points.

    Returns
    -------
    ndata : ndarray
        Array of shifted NMR data.

    """
    return roll(data, pts, neg)


def roll(data, pts=0.0, neg=False):
    """
    Roll axis

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    pts : int
        Number of points to shift.  Positive value will right shift the data,
        negative values will left shift the data.
    neg : bool
        True to negate the shifted points.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with last axis rolled.

    """
    data = np.roll(data, int(pts), axis=-1)
    if neg:
        if pts > 0:
            data[..., :int(pts)] = -data[..., :int(pts)]
        else:
            data[..., int(pts):] = -data[..., int(pts):]
    return data


def fsh(data, pts):
    """
    Frequency shift by Fourier transform. Negative signed phase correction.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    pts : float
        Number of points to frequency shift the data.  Positive value will
        shift the spectrum to the right, negative values to the left.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with last axis rolled.

    """
    s = float(data.shape[-1])

    # inverse fft -> first order phase correction -> fft
    # idata = icomplexft(data)
    # pdata =np.exp(-2.j*pi*pts*np.arange(s)/s,sig=data.dtype)*icomplexft(data)
    # data = complexft(pdata)

    # inplace processing
    return fft(np.exp(-2.j * pi * pts * np.arange(s) /
                      s).astype(data.dtype) * ifft(data))


def fsh2(data, pts):
    """
    Frequency Shift by Fourier transform. Postive signed phase correction.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    pts : float
        Number of points to frequency shift the data.  Positive value will
        shift the spectrum to the right, negative values to the left.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with last axis rolled.

    """
    s = float(data.shape[-1])
    return fft_positive(np.exp(2.j * pi * pts * np.arange(s) /
                        s).astype(data.dtype) * ifft_positive(data))


##############
# Transforms #
##############


def nmr_reorder(data):
    """
    Reorder spectrum after FT transform to NMR order (swap halves and reverse).
    """
    s = data.shape[-1]
    return np.append(data[..., int(s / 2)::-1], data[..., s:int(s / 2):-1],
                     axis=-1)


def swap_halves(data):
    """
    Swap the halves of a spectrum,
    """
    s = data.shape[-1]
    return np.append(data[..., int(s / 2):], data[..., :int(s / 2)], axis=-1)


# Fourier based Transforms
def rft(x):
    """
    Real Fourier transform.

    Parameters
    ----------
    x : ndarray
        Array of NMR data.

    Returns
    -------
    y : ndarray
        Array of NMR data with real Fourier transform applied.

    """
    # XXX figure out what exactly this is doing...
    s = x.shape[-1]
    xp = np.zeros(x.shape, dtype="complex64")
    xp[..., 1:int(s / 2)] = x[..., 1:-1:2] + x[..., 2::2] * 1.j
    xp[..., 0] = x[..., 0] / 2.
    xp[..., int(s / 2)] = x[..., -1] / 2.
    return np.array(nmr_reorder(np.fft.fft(2 * xp, axis=-1).real),
                    dtype="float32")


def irft(xp):
    """
    Inverse real fourier transform

    Parameters
    ----------
    x : ndarray
        Array of NMR data.

    Returns
    -------
    y : ndarray
        Array of NMR data with an inverse real Fourier transform applied.

    """
    # XXX figure out what exactly this is doing
    s = xp.shape[-1]
    xp = np.fft.ifft(nmr_reorder(xp))   # re-order, inverse FT

    # output results
    x = np.zeros(xp.shape, dtype="float32")

    # unpack ifft data
    x[..., 1:-1:2] = xp[..., 1:int(s / 2)].real
    x[..., 2::2] = xp[..., 1:int(s / 2)].imag
    x[..., 0] = xp[..., 0].real
    x[..., -1] = xp[..., int(s / 2)].real
    return x


# Fourier transforms
def fft(data):
    """
    Fourier transform, NMR ordering of results.

    There are a number of definitions of the discrete Fourier transform
    the version used in this function is as follows.

    .. math::

        A_k = \\sum_{m=0}^{n-1} a_m \\exp\\left\\{-2\\pi i{mk \\over n}
        \\right\\}\\qquad k = 0,\\ldots,n-1.

    With the inverse DFT in the :py:func:`ifft` function defined as follows.

    .. math::

        a_m = \\frac{1}{n} \\sum_{k=0}^{n-1} A_k \\exp \\left\\{2\\pi
        i{mk\\over n} \\right\\}\\qquad n = 0,\\ldots,n-1.

    Two alternative definitions are also supported by nmrglue. One in
    which both the sum in the fft and ifft are multiplied by
    :math:`\\frac{1}{\\sqrt{n}}` which results in a pair of transforms in
    which the total power contained in the the signals before and after the
    transforms are equal.  This is the type transforms used in the
    Rowland NMR Toolkit. This type of transform is performed by the
    :py:func:`fft_norm` and :py:func:`ifft_norm` functions.

    The second definition changes the sign of the exponent to be positive while
    keeping the normalization factors the same.  This type of transform is
    performed by the NMRPipe processing package and the functions
    :py:func:`fft_positive` and :py:func:`ifft_positive`.

    All of the Fourier transforms perfromed by nmrglue return results in 'NMR
    order', in which the two half of the spectrum have been swapped and
    reversed.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Fourier transform of NMR data in 'NMR order'.

    See Also
    --------
    ifft : Inversion Fourier transform.
    fft_norm : Norm (power) conserving Fourier transform.
    fft_positive : Forier transform with a positive exponential.


    """
    return np.fft.fftshift(np.fft.fft(data, axis=-1).astype(data.dtype), -1)


def fft_norm(data):
    """
    Fourier transform, total power preserved, NMR ordering of results

    This function is  similar to the transform performed by The Rowland NMR
    Toolkit's FFT function.

    See :py:func:`fft` for documentation of the transformation applied by this
    function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Fourier transform of NMR data in 'NMR order'.

    See Also
    --------
    ifft_norm : Inversion Fourier transform.
    fft : None-norm (power) conserving Fourier transform.
    fft_positive : Forier transform with a positive exponential.

    """
    return fft(data) / np.sqrt(float(data.shape[-1]))


def fft_positive(data):
    """
    Fourier transform with positive exponential, NMR ordering of results

    This is similar to the transform performed by NMRPipe's FFT function.

    See :py:func:`fft` for documentation of the transformation applied by this
    function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Fourier transform of NMR data in 'NMR order'.

    See Also
    --------
    ifft_positive : Inversion Fourier transform.
    fft_norm : Norm (power) conserving Fourier transform.
    fft_positive : Forier transform with a positive exponential.

    """
    # a positive exponential is the same as a IFFT, but we need to undo
    # the 1/N scaling
    s = float(data.shape[-1])
    return (np.fft.fftshift(np.fft.ifft(data, axis=-1).astype(data.dtype), -1)
            * s)


def ifft(data):
    """
    Inverse fourier transform, NMR ordering of results.

    See :py:func:`fft` for documentation of the transformation applied by this
    function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Inverse fourier transform of NMR data in 'NMR order'.

    See Also
    --------
    fft : Fourier transform.
    ifft_norm : Norm (power) conserving inverse Fourier transform.
    ifft_positive : inverse Forier transform with a positive exponential.

    """
    return np.fft.ifft(np.fft.ifftshift(data, -1), axis=-1).astype(data.dtype)


def ifft_norm(data):
    """
    Inverse fourier transform, total power preserved, NMR ordering of results

    This is similar to the transform performed by the Rowland NMR Toolkit's
    IFFT function.

    See :py:func:`fft` for documentation of the transformation applied by this
    function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Inverse fourier transform of NMR data in 'NMR order'.

    See Also
    --------
    fft_norm : Norm (power) conserving Fourier transform.
    ifft : Non-norm (power) conserving inverse Fourier transform.
    ifft_positive : inverse Forier transform with a positive exponential.

    """
    return ifft(data) * np.sqrt(float(data.shape[-1]))


def ifft_positive(data):
    """
    Inverse fourier transform with positive exponential, NMR ordered results.

    This is similar to the transform performed by NMRPipe's FFT function with
    the -inv flag

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Inverse fourier transform of NMR data in 'NMR order'.

    See Also
    --------
    fft_postive : Fourier transform with a positive exponential.
    ifft_norm : Norm (power) conserving inverse Fourier transform.
    ifft : Inverse Forier transform with a negative exponential.

    """
    # a inverse fft with positive exponential in the FFT definition is the
    # same as a FFT with negative exponentials, but with a 1/N scaling factor
    s = 1.0 / float(data.shape[-1])
    return (np.fft.fft(np.fft.ifftshift(data, -1), axis=-1).astype(data.dtype)
            * s)


# Hadamard Transform functions
def int2bin(n, digits=8):
    """
    Integer to binary string
    """
    return "".join([str((n >> y) & 1) for y in range(digits - 1, -1, -1)])


def bin2int(s):
    """
    binary string to integer
    """
    o = 0
    k = len(s) - 1
    for i, v in enumerate(s):
        o = o + int(v) * 2 ** (k - i)
    return o


def gray(n):
    """
    Calculate n-bit gray code
    """
    g = [0, 1]
    for i in range(1, int(n)):
        mg = g + g[::-1]   # mirror the current code
        # first bit 0/2**u for mirror
        first = [0] * 2 ** (i) + [2 ** (i)] * 2 ** (i)
        g = [mg[j] + first[j] for j in range(2 ** (i + 1))]
    return g


def ha(data):
    """
    Hadamard Transform

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Hadamard transform of NMR data.

    Notes
    -----
    This function is very slow.  Implement a Fast Walsh-Hadamard Transform
    with sequency/Walsh ordering (FWHT_w) will result in much faster tranforms.

    http://en.wikipedia.org/wiki/Walsh_matrix
    http://en.wikipedia.org/wiki/Fast_Hadamard_transform

    """
    # implementation is a proof of concept and EXTEMEMLY SLOW

    # determind the order and final size of input vectors
    ord = int(np.ceil(np.log2(data.shape[-1])))  # Walsh/Hadamard order
    max = 2 ** ord

    # zero fill to power of 2
    pad = max - data.shape[-1]
    zdata = zf(data, pad)

    # Multiple each vector by the hadamard matrix
    nat = np.zeros(zdata.shape, dtype=zdata.dtype)
    H = scipy.linalg.hadamard(max)
    nat = np.dot(zdata, H)
    nat = np.array(nat, dtype=data.dtype)

    # Bit-Reversal Permutation
    s = [int2bin(x, digits=ord)[::-1] for x in range(max)]
    brp = [bin2int(x) for x in s]
    brp_data = np.take(nat, brp, axis=-1)

    # Gray code permutation (bit-inverse)
    gp = gray(ord)
    gp_data = np.take(brp_data, gp, axis=-1)

    return gp_data


def ht(data, N=None):
    """
    Hilbert transform.

    Reconstruct imaginary data via hilbert transform.

    Parameters
    ----------
    data : ndarrat
        Array of NMR data.
    N : int or None
        Number of Fourier components.

    Returns
    -------
    ndata : ndarray
        NMR data which has been Hilvert transformed.

    """
    # XXX come back and fix this when a sane version of scipy.signal.hilbert
    # is included with scipy 0.8

    # create an empty output array
    fac = N / data.shape[-1]
    z = np.empty(data.shape, dtype=(data.flat[0] + data.flat[1] * 1.j).dtype)
    if data.ndim == 1:
        z[:] = scipy.signal.hilbert(data.real, N)[:data.shape[-1]] * fac
    else:
        for i, vec in enumerate(data):
            z[i] = scipy.signal.hilbert(vec.real, N)[:data.shape[-1]] * fac

    # correct the real data as sometimes it changes
    z.real = data.real
    return z


##########################
# Standard NMR Functions #
##########################


def di(data):
    """
    Delete imaginary from data

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data without imaginaries.

    """
    return data.real


def ps(data, p0=0.0, p1=0.0, inv=False):
    """
    Linear phase correction

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    p0 : float
        Zero order phase in degrees.
    p1 : float
        First order phase in degrees.
    inv : bool, optional
        True for inverse phase correction

    Returns
    -------
    ndata : ndarray
        Phased NMR data.

    """
    p0 = p0 * pi / 180.  # convert to radians
    p1 = p1 * pi / 180.
    size = data.shape[-1]
    apod = np.exp(1.0j * (p0 + (p1 * np.arange(size) / size))
                  ).astype(data.dtype)
    if inv:
        apod = 1 / apod
    return apod * data


def ps_exp(data, p0=0.0, tc=0.0, inv=False):
    """
    Exponential Phase Correction

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    p0 : float
        Zero order phase in degrees.
    tc : float
        Exponential decay constant.
    inv : bool, optional
        True for inverse phase correction

    Returns
    -------
    ndata : ndarray
        Phased NMR data.

    """
    p0 = p0 * pi / 180.  # convert to radians
    size = data.shape[-1]
    apod = np.exp(1.0j * (p0 * np.exp(-tc * np.arange(size) / size))
                  ).astype(data.dtype)
    if inv:
        apod = 1 / apod
    return apod * data


def tp(data, hyper=False):
    """
    Transpose data.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    hyper : bool
        True if hypercomplex data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with axes transposed.

    """
    if hyper:
        return tp_hyper(data)
    else:
        return data.transpose()


ytp = tp


xy2yx = tp


def tp_hyper(data):
    """
    Hypercomplex tranpose.

    Use when both dimension are complex.

    Parameters
    ----------
    data : ndarray
        Array of hypercomplex NMR data.

    Returns
    -------
    ndata : ndarray
        Array of hypercomplex NMR data with axes transposed.
    """
    return c2ri(ri2c(data).transpose())


def zf_inter(data, pts=1):
    """
    Zero fill between points.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    pts : int
        Number zeros to add between points.

    Returns
    -------
    ndata : ndarray
        Array of NMR data to which `pts` zero have been added between all
        points.

    """
    size = list(data.shape)
    size[-1] = (pts + 1) * size[-1]
    z = np.zeros(size, dtype=data.dtype)
    z[..., ::pts + 1] = data[..., :]
    return z


def zf_pad(data, pad=0, mid=False):
    """
    Zero fill by padding with zeros.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    pad : int
        Number of zeros to pad data with.
    mid : bool
        True to zero fill in middle of data.

    Returns
    -------
    ndata : ndarray
        Array of NMR data to which `pad` zeros have been appended to the end or
        middle of the data.

    """
    size = list(data.shape)
    size[-1] = int(pad)
    z = np.zeros(size, dtype=data.dtype)

    if mid:
        h = int(data.shape[-1] / 2.0)
        return np.concatenate((data[..., :h], z, data[..., h:]), axis=-1)
    else:
        return np.concatenate((data, z), axis=-1)


zf = zf_pad


def zf_double(data, n, mid=False):
    """
    Zero fill by doubling original data size once or multiple times.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    n : int
        Number of times to double the size of the data.
    mid : bool
        True to zero fill in the middle of data.

    Returns
    -------
    ndata : ndarray
        Zero filled array of NMR data.

    """
    return zf_pad(data, int((data.shape[-1] * 2 ** n) - data.shape[-1]), mid)


def zf_size(data, size, mid=False):
    """
    Zero fill to given size.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    size : int
        Size of data after zero filling.
    mid : bool
        True to zero fill in the middle of data.

    Returns
    -------
    ndata : ndarray
        Zero filled array of NMR data.

    """
    return zf_pad(data, int(size - data.shape[-1]), mid)


def largest_power_of_2(value):
    """
    Find the nearest power of two equal to or larger than a value.

    Parameters
    ----------
    value : int
        Value to find nearest power of two equal to or larger than.

    Returns
    -------
    pw : int
        Power of 2.

    """
    return int(pow(2, np.ceil(np.log(value) / np.log(2))))


def zf_auto(data, mid=False):
    """
    Zero fill to next largest power of two.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    mid : bool
        True to zero fill in the middle of data.

    Returns
    -------
    ndata : ndarray
        Zero filled array of NMR data.

    """
    return zf_size(data, largest_power_of_2(data.shape[-1]), mid)


####################
# Basic Untilities #
####################


# Add Constant
def add(data, r=0.0, i=0.0, c=0.0):
    """
    Add constant.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    r : float
        Constant to add to real data.
    i : float
        Constant to add to imaginary data, ignored if no data does not have
        imaginary data.
    c : float
        Constant to add to both real and imaginary data.  This constant is only
        added to the real data if no imaginary data is present.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with constant added.

    """
    data.real = data.real + r + c
    if np.iscomplexobj(data):
        data.imag = data.imag + i + c
    return data


def add_ri(data):
    """
    Add real and imaginary components.

    Parameters
    ----------
    data : ndarray
        Array of complex NMR data.

    Returns
    -------
    ndata : ndarray
        Sum of real and imaginary component of NMR data.

    """
    return data.real + data.imag


# Derivative
def dx(data):
    """
    Derivative by central difference

    Edges are takes as difference between nearest points

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Derivate of NMR data.

    """
    z = np.empty_like(data)
    z[..., 0] = data[..., 1] - data[..., 0]    # first point
    z[..., -1] = data[..., -1] - data[..., -2]  # last point
    z[..., 1:-1] = data[..., 2:] - data[..., :-2]  # interior
    return z


# Extract Region
def ext(data, x0=None, xn=None, y0=None, yn=None):
    """
    Extract region.

    Data should be two dimensional. Axes are labeled Y, X.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    x0 : int
        X-axis extract region start.
    xn : int
        X-axis extract region end.
    y0 : int
        Y-axis extract region start.
    yn : int
        Y-axis extract region end.

    Returns
    -------
    ndata : ndarray
        Extracted region.

    """
    return data[y0:yn, x0:xn]


def ext_left(data):
    """
    Extract the left half of spectrum.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Left half of spectrum.

    """
    return data[..., 0:int(data.shape[-1] / 2.)]


def ext_right(data):
    """
    Extract the right half of the spectrum.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Right half of spectrum.

    """
    return data[..., int(data.shape[-1] / 2.):]


def ext_mid(data):
    """
    Extract the middle of the spectrum.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Middle half of spectrum.

    """
    return data[..., int(data.shape[-1] * 1. / 4.):
                int(data.shape[-1] * 3. / 4.)]


# Integrate
def integ(data):
    """
    Integrate by cumulative sum. The units of the integral is in points (i.e.
    dx is equal to 1).

    Parameters
    ----------
    data : ndarray
        Array of NMR data.

    Returns
    -------
    ndata : ndarray
        Integrated NMR data.

    """
    return np.cumsum(data, axis=-1)


# Modulus Calculation
def mc(data):
    """
    Modulus calculation.

    Calculates sqrt(real^2 + imag^2)
    """
    return np.sqrt(data.real ** 2 + data.imag ** 2)


def mc_pow(data):
    """
    Modulus calculation. Squared version.

    Calculated real^2+imag^2
    """
    return data.real ** 2 + data.imag ** 2


# Mirror
def mir_left(data):
    """
    Append a mirror image of the data on the left.
    """
    return np.append(data, data[..., ::-1], axis=-1)


def mir_right(data):
    """
    Append a mirror image of the data on the right.
    """
    return np.append(data[..., ::-1], data, axis=-1)


def mir_center(data):
    """
    Append a mirror image of the data in the center.
    """
    s = data.shape[-1]
    return np.concatenate(
        (data[..., int(s / 2):], data, data[..., :int(s / 2)]), axis=-1)


def mir_center_onepoint(data):
    """
    Append a mirror image of the data in the center with a one point shift
    amd negate appended imaginary data.
    """
    s = int(data.shape[-1])
    data = np.concatenate((data[..., s - 1:0:-1], data), axis=-1)
    if np.iscomplexobj(data):
        data.imag[..., :s - 1] = -data.imag[..., :s - 1]
    return data


# Multiply by a constant
def mult(data, r=1.0, i=1.0, c=1.0):
    """
    Multiply by a constant.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    r : float
        Constant to multiply real channel by.
    i : float
        Constant to multiply imaginary channel by. If data has no imaginary
        component, this parameter is ignored.
    c : float
        Constant to multiply both channels by.

    Returns
    -------
    ndata : ndarray
        Array of NMR data which has been multiplied by a constant.

    """
    data.real = data.real * r * c
    if np.iscomplexobj(data):
        data.imag = data.imag * i * c
    return data


# Reverse
def rev(data):
    """
    Reverse data.
    """
    return data[..., ::-1]


# Set to a Constant
def set(data, c):
    """
    Set to a constant.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    c : float or complex
        Constant to set data to.

    Returns
    -------
    ndata : ndarray
        Array of constants.

    Notes
    -----
    data is set in place, if original data is required make a copy before
    calling this function.

    """
    data[..., :] = c
    return data


def set_complex(data, v):
    """
    Set real and imaginary portions of data to a constant.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    v : float
        Constant to set both real and imaginary component of `data` to.

    Returns
    -------
    ndata : ndarray
        Array with real and imaginary components set to v.

    Notes
    -----
    data is set in place, if original data is required make a copy before
    calling this function.

    """
    data.real = v
    if np.iscomplexobj(data):
        data.imag = v
    return data


def set_real(data, v):
    """
    Set real component of data to a constant.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    v : float
        Constant to set both real component of `data` to.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with real components set to v.

    Notes
    -----
    data is set in place, if original data is required make a copy before
    calling this function.

    """
    data.real = v
    return data


def set_imag(data, v):
    """
    Set imaginary portion of data to a constant.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    v : float
        Constant to set both imaginary component of `data` to.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with imaginary components set to v.

    Notes
    -----
    data is set in place, if original data is required make a copy before
    calling this function.

    """
    if np.iscomplexobj(data):
        data.imag = v
    return data


# Shuffle Utilities
def ri2c(data):
    """
    Interleave real and imaginary data into a real array.
    """
    s = list(data.shape)
    s[-1] = s[-1] * 2
    n = np.empty(s, data.real.dtype)
    n[..., ::2] = data.real
    n[..., 1::2] = data.imag
    return n


def interleave_complex(data):
    """
    Unpack complex data into an interleaved real, imaginary array.
    """
    return ri2c(data)


def unpack_complex(data):
    """
    Unpacks complex array into real array (interleaves values).
    """
    return ri2c(data)


def c2ri(data):
    """
    Seperate interleaved real, imaginary data into complex array.

    Assumes data is real only, ignores imaginary portion of data.

    """
    # make a 1,1 array to determind dtype
    temp = np.array(data.flat[0] + data.flat[1] * 1j)
    s = list(data.shape)
    s[-1] = int(s[-1] / 2)
    n = np.empty(s, temp.dtype)
    del(temp)
    n.real = data.real[..., ::2]
    n.imag = data.real[..., 1::2]
    return n


def seperate_interleaved(data):
    """
    Seperate interleaved real, imaginary data into complex array.
    """
    return c2ri(data)


def pack_complex(data):
    """
    Packs interleaved real array into complex array.
    """
    return c2ri(data)


def decode_States(data):
    """
    Decode data collected using States (seperates interleaved data).
    """
    return c2ri(data)


def ri2rr(data):
    """
    Append imaginary data to end of real data, returning a real array.
    """
    s = list(data.shape)
    half = int(s[-1])
    s[-1] = half * 2
    n = np.empty(s, data.real.dtype)
    n[..., :half] = data.real
    n[..., half:] = data.imag
    return n


append_imag = ri2rr


def rr2ri(data):
    """
    Unappend real and imaginary data returning a complex array.
    """
    # make a 1,1 array to determind dtype
    temp = np.array(data.flat[0] + data.flat[1] * 1.j)
    s = list(data.shape)
    half = int(s[-1] / 2.0)
    s[-1] = half
    n = np.empty(s, temp.dtype)
    del(temp)
    n.real = data[..., :half]
    n.imag = data[..., half:]
    return n


unappend_imag = rr2ri


def exlr(data):
    """
    Exchange left and right halves of array.
    """
    half = int(data.shape[-1] / 2)
    n = np.empty_like(data)
    n[..., :half] = data[..., half:]
    n[..., half:] = data[..., :half]
    return n


exchange_lr = exlr


def rolr(data):
    """
    Rotate left and right halves of array.
    """
    half = int(data.shape[-1] / 2)
    n = np.empty_like(data)
    n[..., :half] = data[..., (half - 1)::-1]
    n[..., half:] = data[..., :(half - 1):-1]
    return n


rotate_lr = rolr


def swap(data):
    """
    Swap real and imaginary data.
    """
    n = np.empty_like(data)
    n.real = data.imag
    n.imag = data.real
    return n


swap_ri = swap


def bswap(data):
    """
    Byteswap data
    """
    return data.byteswap()


byte_swap = bswap


# Sign Manipulation Utilities
def neg_left(data):
    """
    Negate left half.
    """
    data[..., 0:int(data.shape[-1] / 2.)] = \
        -data[..., 0:int(data.shape[-1] / 2.)]
    return data


def neg_right(data):
    """
    Negate right half.
    """
    data[..., int(data.shape[-1] / 2.):] = \
        -data[..., int(data.shape[-1] / 2.):]
    return data


def neg_middle(data):
    """
    Negate middle half.
    """
    data[..., int(data.shape[-1] * 1. / 4.):int(data.shape[-1] * 3. / 4.)] = \
        -data[..., int(data.shape[-1] * 1. / 4.):int(data.shape[-1] * 3. / 4.)]
    return data


def neg_edges(data):
    """
    Negate edge half (non-middle) of spectra.
    """
    data[..., :int(data.shape[-1] * 1. / 4)] = \
        -data[..., :int(data.shape[-1] * 1. / 4)]
    data[..., int(data.shape[-1] * 3. / 4):] = \
        -data[..., int(data.shape[-1] * 3. / 4):]
    return data


def neg_all(data):
    """
    Negate data
    """
    return -data


def neg_real(data):
    """
    Negate real data
    """
    data.real = -data.real
    return data


def neg_imag(data):
    """
    Negate imaginary data
    """
    data.imag = -data.imag
    return data


def neg_even(data):
    """
    Negate even points
    """
    data[..., ::2] = -data[..., ::2]
    return data


def neg_odd(data):
    """
    Negate odd points
    """
    data[..., 1::2] = -data[..., 1::2]
    return data


def neg_alt(data):
    """
    Negate alternate (odd) points.
    """
    return neg_odd(data)


def abs(data):
    """
    Replace data with absolute value of data (abs of real, imag seperately)
    """
    data.real = np.abs(data.real)
    data.imag = np.abs(data.imag)
    return data


def sign(data):
    """
    Replace data with sign (-1 or 1) of data (seperately on each channel)
    """
    data.real = np.sign(data.real)
    data.imag = np.sign(data.imag)
    return data


##################
# Misc Functions #
##################


# Coadd data
def coadd(data, clist, axis=-1):
    """
    Coadd data.

    Reduce data along axis by blocks data and multiplying by
    coefficients in clist.  Incomplete blocks are discarded.

    Parameters
    ----------
    data : 2D ndarray
        Array of NMR data data.
    clist : list
        List of Coefficients
    axis : {0, 1, -1}
        Axis to reduce. 0 for Y, 1 or -1 for X.

    """
    # there is probably a efficient way to do this with tile and inner
    # or scipy.ndimage.generic_filter

    # algorith creates a empty array, then fills it element wise
    # with each factor from clist times the blocks selected

    s = list(data.shape)    # data shape
    k = len(clist)          # length of coefficient list

    if axis == 1 or axis == -1:   # 'x' axis
        s[-1] = int(np.floor(float(s[-1]) / k))
        n = np.zeros(s, dtype=data.dtype)
        m = s[-1] * k   # last element read
        for i in range(k):
            n = n + clist[i] * data[..., i:m:k]
    else:   # 'y' axis
        s[0] = int(np.floor(float(s[0]) / k))
        n = np.zeros(s, dtype=data.dtype)
        m = s[0] * k
        for i in range(k):
            n = n + clist[i] * data[i:m:k]
    return n


coad = coadd


# Image Processing
def thres(data, thres=0.0):
    """
    Mark values less than thres as invalid (for use with filters)

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    thres : float
        Threshold value.

    Returns
    -------
    ndata : masked ndarray
        Masked array of NMR data with values less than thres masked.

    """
    return np.ma.masked_less(data, thres)


def conv(data, kern=[1.], m="wrap", c=0.0):
    """
    Convolute data with kernel.

    Real and imaginary components of data are convolved seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    kern : array_like
        Convolution kernel
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of convolution are dealt with.
    c : float
        Constant value for use when m is  'constant'.

    Returns
    -------
    ndata : ndarray
        NMR data which has been convolved with the given kernel.

    """
    kern = np.array(kern)
    data.real = scipy.ndimage.convolve(data.real, weights=kern, mode=m, cval=c)
    data.imag = scipy.ndimage.convolve(data.imag, weights=kern, mode=m, cval=c)
    return data


convolute = conv


def corr(data, kern=[1.], m="wrap", c=0.0):
    """
    Correlate data with a kernel (weights).

    Real and imaginary components of data are correlated seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    kern : array_like
        Correlation kernel (weights).
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of correlation are dealt with.
    c : float
        Constant value for use when m is  'constant'.

    Returns
    -------
    ndata : ndarray
        NMR data which has been correlated with the given kernel.

    """
    kern = np.array(kern)
    data.real = scipy.ndimage.correlate(data.real, weights=kern, mode=m,
                                        cval=c)
    data.imag = scipy.ndimage.correlate(data.imag, weights=kern, mode=m,
                                        cval=c)
    return data


correlate = corr


def filter_median(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a median filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    data.real = scipy.ndimage.median_filter(data.real, size=s, mode=m, cval=c)
    data.imag = scipy.ndimage.median_filter(data.imag, size=s, mode=m, cval=c)
    return data


def filter_min(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a minimum filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    data.real = scipy.ndimage.minimum_filter(data.real, size=s, mode=m, cval=c)
    data.imag = scipy.ndimage.minimum_filter(data.imag, size=s, mode=m, cval=c)
    return data


def filter_max(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a maximum filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    data.real = scipy.ndimage.maximum_filter(data.real, size=s, mode=m, cval=c)
    data.imag = scipy.ndimage.maximum_filter(data.imag, size=s, mode=m, cval=c)
    return data


def filter_percentile(data, percentile, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a percentile filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    percentile : float
        Filter percentile parameter.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    data.real = scipy.ndimage.percentile_filter(data.real, percentile, size=s,
                                                mode=m, cval=c)
    data.imag = scipy.ndimage.percentile_filter(data.imag, percentile, size=s,
                                                mode=m, cval=c)
    return data


def filter_rank(data, rank, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a rank filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    rank : int
        Filter rank parameter.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    data.real = scipy.ndimage.rank_filter(data.real, rank, size=s, mode=m,
                                          cval=c)
    data.imag = scipy.ndimage.rank_filter(data.imag, rank, size=s, mode=m,
                                          cval=c)
    return data


# These filter are much slower as they use the generic filter functions...


def filter_amin(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply an absolute minimum filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = amin_flt
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


def filter_amax(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply an absolute maximum filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = amax_flt
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


def filter_range(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a range filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = range_flt
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


def filter_avg(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply an average filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = avg_flt
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


def filter_dev(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a standard deviation filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = std_flt
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


def filter_sum(data, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a summation filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = sum_flt
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


def filter_generic(data, filter, s=(1, 1), m="wrap", c=0.0):
    """
    Apply a generic filter.

    Real and imaginary components are filtered seperately.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    filter : functions
        Python function which takes an array and returns a single value.
    s : tuple
        Shape or size taken for each step of the filter.
    m : {'reflect', 'constant', 'nearest', 'wrap'}
        How edges of are dealt with.
    c : float
        Constant value for use when m is 'constant'.

    Returns
    -------
    ndata : ndarray
        Filtered NMR data.

    """
    flt = filter
    data.real = scipy.ndimage.generic_filter(data.real, flt, size=s, mode=m,
                                             cval=c)
    data.imag = scipy.ndimage.generic_filter(data.imag, flt, size=s, mode=m,
                                             cval=c)
    return data


# filter functions

def amin_flt(arr):
    return arr[np.abs(arr).argmin()]


def amax_flt(arr):
    return arr[np.abs(arr).argmax()]


def range_flt(arr):
    return arr.max() - arr.min()


def avg_flt(arr):
    return arr.avg()


def std_flt(arr):
    return arr.std()


def sum_flt(arr):
    return arr.sum()


# Scale Quad Artifacts
def qart(data, a=0.0, f=0.0):
    """
    Scale Quad Artifacts.

    Replaces imaginary data with (1 + a) * data.imag + f * data.real

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    a : float
        Amplitude adjustment.
    f : float
        Phase adjustment.

    Returns
    -------
    ndata: ndarray
        Array of NMR data with quadrature artifacts scaled.

    """
    data.imag = (1 + a) * data.imag + f * data.real
    return data


def qart_auto(data):
    """
    Scale quad artifacts by values from Gram-Schmidt orthogonalization.
    """
    a, f = gram_schmidt(data)
    return qart(data, a, f)


def gram_schmidt(data):
    """
    Calculate Gram-Schmidt orthogonalization parameters.
    """
    # method similar to Hock and Stern, "NMR Data Processing" p.61
    # sum of correlation between data.real and data.imag
    C = (data.real * data.imag).sum()

    # total power in real channel
    R = (data.real * data.real).sum()

    # remove correlation from imag channel
    idata = data.imag - (C / R) * data.real

    # total power in uncorrelated imag channel
    S = (idata * idata).sum()

    # imag(data'') = R/S*imag(data')
    # imag(data')  = imag(data)-C/R * real(data)
    # therefore:
    # imag(data'') = R/S*imag(data) - R*C/(S*R) * real(data)
    # so A = R/S, B=-C/(S)
    return(R / S, -C / S)


# Complex Mixing
def qmix(data, carr):
    """
    Mix input and output channels provided coefficient array.

    Parameters
    ----------
    data : 2D ndarray
        Array of NMR data.
    carr : array_liek
        Array of coefficients for mixing.  The size of carr must evenly divide
        qmix.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with channels mixed.

    """
    carr = np.array(carr).transpose()
    ic = carr.shape[1]  # input channels
    oc = carr.shape[0]  # output channels

    if data.shape[0] % ic != 0 or data.shape[0] % oc != 0:
        raise ValueError("Coefficient array does not evenly divide data")

    # create an empty blank output array
    s = list(data.shape)
    s[0] = int(s[0] * float(oc) / float(ic))
    n = np.empty(s, data.dtype)

    # remix each block
    for i in range(int(data.shape[0] / float(ic))):
        block = data[i * ic:(i + 1) * ic]
        n[i * oc:(i + 1) * oc] = np.dot(carr, block)

    return n


# Smooth and Center
def smo(data, n):
    """
    Smooth data.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    n : int
        Size of smoothing window (+/- points)

    Returns
    -------
    ndata : ndarray
        Array of smoothed NMR data.

    """
    # XXX this can probably be accomplished by a median_filter
    n = int(n)

    # a is the accumulator
    a = np.copy(data)

    for i in range(1, n + 1):
        a = a + rs(data, i) + ls(data, i)

    # divide the interior by 2*n+1 to get mean
    a[..., n:-n] = a[..., n:-n] / (2 * n + 1)
    # divide the left edges by 2n+1-i where i is the distance from the interior
    for i in range(1, n + 1):
        a[..., n - i] = a[..., n - i] / (2. * n + 1 - i)
    # divide the right edge similarly
    for i in range(-n, 0):
        a[..., i] = a[..., i] / (n - i)
    return a


def center(data, n):
    """
    Center data.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    n : int
        Size of centering window (+/- points)

    Returns
    -------
    ndata : ndarray
        Array of centered NMR data.

    """
    return data - smo(data, n)


def zd(data, window, x0=0.0, slope=1.0):
    """
    Zero Diagonal band with generic window function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    window : ndarray
        Window to apply to diagonal band.
    wide : int
        Diagonal band half width in points.
    x0 : int
        Starting location of diagonal band in points.
    slope : float
        Slope of diagonal band.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with diagonal band set to zero.


    """
    width = len(window)             # full width
    wide = int((width - 1.) / 2)    # half width
    rows = data.shape[0]            # rows in data
    cols = data.shape[-1]           # columns in data
    c_start = x0 + slope            # start of center diagonal band

    # last row to apply window to is last row or where we run off the grid
    max_r = int(min(rows, np.floor((cols - c_start + wide) / slope) + 1))

    # apply window to band row by row
    for r in range(max_r):  # r from 0 to max_r-1
        w_min = 0           # window min
        w_max = width       # window max

        c_mid = int(r * slope + (c_start))  # middle of diagonal band
        c_min = c_mid - wide
        c_max = c_mid + wide + 1

        if c_min < 0:
            w_min = int(-c_min)
            c_min = 0
        if c_max > cols:
            w_max = int(w_max - (c_max - cols))
            c_max = cols

        data[r, c_min:c_max] = data[r, c_min:c_max] * window[w_min:w_max]

    return data


def zd_boxcar(data, wide=1, x0=0.0, slope=1.0):
    """
    Zero diagonal band with a boxcar function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    wide : int
        Diagonal band half width in points.
    x0 : int
        Starting location of diagonal band in points.
    slope : float
        Slope of diagonal band.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with diagonal band set to zero.

    """
    window = np.zeros(2 * int(wide) + 1)
    return zd(data, window, x0=x0, slope=slope)


def zd_triangle(data, wide=1.0, x0=0.0, slope=1.0):
    """
    Zero diagonal band with triangle function

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    wide : int
        Diagonal band half width in points.
    x0 : int
        Starting location of diagonal band in points.
    slope : float
        Slope of diagonal band.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with diagonal band set to zero.

    """
    window = np.append(np.linspace(1, 0, wide + 1),
                       np.linspace(0, 1, wide + 1)[1:])
    return zd(data, window, x0=x0, slope=slope)


def zd_sinebell(data, wide=1.0, x0=0.0, slope=1.0):
    """
    Zero diagonal band with sinebell function.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    wide : int
        Diagonal band half width in points.
    x0 : int
        Starting location of diagonal band in points.
    slope : float
        Slope of diagonal band.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with diagonal band set to zero.

    """
    window = 1 - np.sin(np.linspace(0, pi, 2 * wide + 1))
    return zd(data, window, x0=x0, slope=slope)


def zd_gaussian(data, wide=1.0, x0=0.0, slope=1.0, g=1):
    """
    Zero Diagonal band with gaussian function

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    wide : int
        Diagonal band half width in points.
    x0 : int
        Starting location of diagonal band in points.
    slope : float
        Slope of diagonal band.
    g : float
        Width of Gaussian function.

    Returns
    -------
    ndata : ndarray
        Array of NMR data with diagonal band set to zero.

    """
    tln2 = np.sqrt(2 * np.log(2))
    window = 1 - scipy.signal.gaussian(2 * wide + 1, g / tln2)
    return zd(data, window, x0=x0, slope=slope)



"""
Automated phase correction
These functions provide support for automatic phasing of NMR data. They
consist of the core `autops` function which performs the optimisation and
a set of private functions for calculating a spectral phase quality score
for a provided spectrum.
"""

from __future__ import print_function

import numpy as np
import scipy.optimize

from .proc_base import ps


def autops(data, fn, p0=0.0, p1=0.0, return_phases=False, peak_width=100, **kwargs):
    """
    Automatic linear phase correction

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    fn : str or function
        Algorithm to use for phase scoring. Built in functions can be
        specified by one of the following strings: "acme", "peak_minima"
    p0 : float
        Initial zero order phase in degrees.
    p1 : float
        Initial first order phase in degrees.
    peak_width : int
        Width of the ROI for peak_minima optimization (number of surrounding indexes)
    return_phases : Bool
        returns a list of optimized values of phases [p0, p1] in addition
        to the phased data
    kwargs : additional key-word arguments to be passed to the solver
        The are documented at the following link:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.fmin.html
        Some of the more useful ones for this use case:

        * disp : Bool
            Turns on or off the printing of convergence messeges
            By default, this is set to True.
        * ftol : float
            Absolute error in fn between iterations that is acceptable for
            convergence. The default is 1e-4.

    Returns
    -------
    ndata : ndarray
        Phased NMR data.

    """
    arguments = [data,]

    if not callable(fn):
        if fn == 'peak_minima':
            arguments.append(peak_width)

        try:
            fn = {
                'peak_minima': _ps_peak_minima_score,
                'acme': _ps_acme_score,
            }[fn]
        except KeyError:
            raise KeyError('Unable to find algorithm "{}". Use "acme", "peak_minima" or pass a custom function.'.format(fn))

    opt = [p0, p1]

    opt = scipy.optimize.fmin(fn, x0=opt, args=tuple(arguments), **kwargs)

    phasedspc = ps(data, p0=opt[0], p1=opt[1])

    if return_phases:
        return phasedspc, opt
    else:
        return phasedspc


def _ps_acme_score(ph, data):
    """
    Phase correction using ACME algorithm by Chen Li et al.
    Journal of Magnetic Resonance 158 (2002) 164-168

    Parameters
    ----------
    ph : tuple
        Current p0 and p1 values
    data : ndarray
        Array of NMR data.

    Returns
    -------
    score : float
        Value of the objective function (phase score)

    """
    stepsize = 1

    phc0, phc1 = ph

    s0 = ps(data, p0=phc0, p1=phc1)
    data = np.real(s0)

    # Calculation of first derivatives
    ds1 = np.abs((data[1:]-data[:-1]) / (stepsize*2))
    p1 = ds1 / np.sum(ds1)

    # Calculation of entropy
    p1[p1 == 0] = 1

    h1 = -p1 * np.log(p1)
    h1s = np.sum(h1)

    # Calculation of penalty
    pfun = 0.0
    as_ = data - np.abs(data)
    sumas = np.sum(as_)

    if sumas < 0:
        pfun = pfun + np.sum((as_/2) ** 2)

    p = 1000 * pfun

    return (h1s + p) / data.shape[-1] / np.max(data)


def _ps_peak_minima_score(ph, data, peak_width):
    """
    Phase correction using simple minima-minimisation around highest peak

    This is a naive approach but is quick and often achieves reasonable
    results.  The optimisation is performed by finding the highest peak in the
    spectra (e.g. TMSP) and then attempting to reduce minima surrounding it.

    Parameters
    ----------
    ph : tuple
        Current p0 and p1 values
    peak_width : int
        Lookup width
    data : ndarray
        Array of NMR data.

    Returns
    -------
    score : float
        Value of the objective function (phase score)

    """

    phc0, phc1 = ph

    s0 = ps(data, p0=phc0, p1=phc1)
    data = np.real(s0)

    i = np.argmax(data)
    mina = np.min(data[i-peak_width:i])
    minb = np.min(data[i:i+peak_width])

    return np.abs(mina - minb)


def manual_ps(data, notebook=False):
    """
    Manual Phase correction using matplotlib

    A matplotlib widget is used to manually correct the phase of a Fourier
    transformed dataset. If the dataset has more than 1 dimensions, the first
    trace will be picked up for phase correction.  Clicking the 'Set Phase'
    button will print the current linear phase parameters to the console.
    A ipywidget is provided for use with Jupyter Notebook to avoid changing
    backends. This can be accessed with notebook=True option in this function

    .. note:: Needs matplotlib with an interactive backend.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    notebook : Bool
        True for plotting interactively in Jupyter Notebook
        Uses ipywidgets instead of matplotlib widgets

    Returns
    -------
    p0, p1 : float
        Linear phase correction parameters. Zero and first order phase
        corrections in degrees calculated from pc0, pc1 and pivot displayed
        in the interactive window.

    Examples
    --------
    >>> import nmrglue as ng
    >>> p0, p1 = ng.process.proc_autophase.manual_ps(data)
    >>> # do manual phase correction and close window
    >>> phased_data = ng.proc_base.ps(data, p0=p0, p1=p1)

    If you are using a Jupyter Notebook::

        In  [1] ng.process.proc_autophase.manual_ps(data)
        Out [1] # do manual phase correction. p0 and p1 values will be updated
                # continuously as you do so and are printed below the plot
        In  [2] phased_data = ng.proc_base.ps(data, p0=p0, p1=p1)

    """

    if len(data.shape) == 2:
        data = data[0, ...]
    elif len(data.shape) == 3:
        data = data[0, 0, ...]
    elif len(data.shape) == 4:
        data = data[0, 0, 0, ...]

    if notebook:
        from ipywidgets import interact, fixed
        import matplotlib.pyplot as plt

        def phasecorr(dataset, phcorr0, phcorr1, pivot):
            fig, ax = plt.subplots(figsize=(10, 7))
            phaseddata = dataset * np.exp(
                1j * (phcorr0 + phcorr1 * (
                    np.arange(-pivot, -pivot+dataset.size)/dataset.size)))

            ax.plot(np.real(phaseddata))
            ax.set(ylim=(np.min(np.real(data))*2, np.max(np.real(data))*2))
            ax.axvline(pivot, color='r', alpha=0.5)
            plt.show()

            p0 = np.round(
                (phcorr0 - phcorr1 * pivot/dataset.size) * 360 / 2 / np.pi, 3)
            p1 = np.round(phcorr1*360/2/np.pi, 3)

            print('p0 =', p0, 'p1 =', p1)

        interact(
            phasecorr,
            dataset=fixed(data),
            phcorr0=(-np.pi, np.pi, 0.01),
            phcorr1=(-10*np.pi, 10*np.pi, 0.01),
            pivot=(0, data.size, 1))

    else:

        from matplotlib.widgets import Slider, Button
        import matplotlib.pyplot as plt

        plt.subplots_adjust(left=0.25, bottom=0.35)

        interactive, = plt.plot(data.real, lw=1, color='black')

        axcolor = 'white'
        axpc0 = plt.axes([0.25, 0.10, 0.65, 0.03], facecolor=axcolor)
        axpc1 = plt.axes([0.25, 0.15, 0.65, 0.03], facecolor=axcolor)
        axpiv = plt.axes([0.25, 0.20, 0.65, 0.03], facecolor=axcolor)
        axpst = plt.axes([0.25, 0.25, 0.15, 0.04], facecolor=axcolor)

        spc0 = Slider(axpc0, 'p0', -360, 360, valinit=0)
        spc1 = Slider(axpc1, 'p1', -360, 360, valinit=0)
        spiv = Slider(axpiv, 'pivot', 0, data.size, valinit=0)
        axps = Button(axpst, 'Set Phase', color=axcolor)

        def update(val):
            pc0 = spc0.val * np.pi / 180
            pc1 = spc1.val * np.pi / 180
            pivot = spiv.val
            interactive.set_ydata((data * np.exp(
                1.0j * (pc0 + (pc1 * np.arange(-pivot, -pivot + data.size) /
                        data.size))).astype(data.dtype)).real)
            plt.draw()

        def setphase(val):
            p0 = spc0.val-spc1.val*spiv.val/data.size
            p1 = spc1.val
            print(p0, p1)

        spc0.on_changed(update)
        spc1.on_changed(update)
        spiv.on_changed(update)
        axps.on_clicked(setphase)

        plt.show(block=True)

        p0 = spc0.val-spc1.val*spiv.val/data.size
        p1 = spc1.val
        return p0, p1



"""
xcpy - interfaces with external programs from within Bruker-TopSpin

USAGE
xcpy
xcpy [OPTIONS]
xcpy [OPTIONS] [SCRIPTNAME]

INSTALLATION
1. Copy (or symlink) this file to the following directory:
<topspin_location>/exp/stan/nmr/py/user/
2. If you now type 'xcpy' on the command line within Topspin,
this documentation should pop up
3. A configuration file needs to be written out so that xcpy
knows the location of the CPython executable and a folder where
the .py scripts are located. This can be done by 'xcpy -s'. This
can be rewritten at any time point using the same command.


DESCRIPTION
xcpy supports running external scripts via Jython (subprocess module)
that ships with Topspin. Currently, it allows only external CPython
programs to run. By default, it passes the current folder, expno and procno
to the external CPython program (if available).


OPTIONS
-h, --help: Brings up this docstring. Also brings it up when no other
\toption is given.

-s, --settings: Opens a dialog to write out a configuration file. The
\tlocation of the Cpython executable and the folder where all scripts
\tare located can be given. Use full path names for this, i.e., use
\t'/usr/bin/python3' for *nix instead of 'python3' or '~/folder/python3',
\tand 'C:\python.exe' for Windows (note the .exe) extension. If the
\tconfiguration file already exists, the entries will be verified and
\tthe dialog will be populated with them if these are found to be correct.
\tElse, an error message with the configuration file will be returned and
\tthe dialogs will be kept empty.

-n, --name: Opens a dialog box to give the path of a script to be run

-c, --config: Prints contents of the configuration file, if it exists.
\tIf no configuration file is found, it prints 'None'

-d, --dry-run: Prints out the command that will be executed by the subprocess
\tmodule if this option is not given.

--no-args: Does not pass any arguments (current data folder, etc) to
\tthe external program

--use-shell: Uses shell to run the subprocess command. By default, this is
\tnot used for *nix, but is used for Windows. [Warning: this is a known
\tsecurity risk and users are advised to be careful with their input]

"""
import sys
import os
from subprocess import Popen, PIPE, STDOUT

try:
    from ConfigParser import SafeConfigParser
except ModuleNotFoundError:
    # this will only fail if Python 3 is used
    # but that error willl be handled by the
    # check_jython() function
    pass


def topspin_error():
    errmsg = """
    This file is meant to be executed
    using Jython from within Topspin
    Please see the doctring for more
    details.
    """
    return errmsg


def check_jython():
    """
    Checks whether Jython is being used to run this script
    from within Topspin

    """
    # some of the functions defined in the global namespace
    # by Topspin which are also used by xcpy
    topspin_inbuilts = ["MSG", "INPUT_DIALOG", "CURDATA"]

    g = globals().keys()
    for function in topspin_inbuilts:
        if function in g:
            pass
        else:
            raise Exception(topspin_error())
    return True


def topspin_location():
    """
    Gets Topspin home directory. Also serves to check
    whether the script is being executed from within
    Topspin or externally.

    """
    try:
        # topspin >= 3.6
        toppath = sys.getBaseProperties()["XWINNMRHOME"]
    except AttributeError:
        try:
            # topspin >= 3.1 and <= 3.5
            toppath = sys.registry["XWINNMRHOME"]
        except (AttributeError, KeyError):
            # topspin < 3.1
            toppath = sys.getEnviron()["XWINNMRHOME"]

    # if all the above fail, there should be an exception raised and
    # the function should not return anything
    return toppath


def read_cfg(filename, check_python=True):
    """
    Reads in the configuration file

    """
    config = SafeConfigParser()
    config.read(filename)

    try:
        cpyname = config.get("xcpy", "cpython")
    except (NoSectionError, NoOptionerror, KeyError):
        cpyname = ""

    try:
        scripts_location = config.get("xcpy", "scripts_location")
        scripts_location = scripts_location.split(",")
    except (NoSectionError, NoOptionerror, KeyError):
        scripts_location = [""]

    if check_python:
        if cpyname:
            verify_python(cpyname)

    return cpyname, scripts_location


def write_cfg(outfile, infile=None):
    """
    Writes or overwrites a configuration file

    """
    if infile is not None:
        try:
            cpyname, scripts_location = read_cfg(infile)
        except OSError:
            if exists(infile):
                errmsg = """
                    The following configuration was found in the file {}:

                    {}

                    These settings are likely incorrect.
                    You can enter the correct settings at the next dialog box.
                    Press 'Close' to continue.
                    """.format(
                    infile, show_config(infile, printing=False)
                )
                MSG(errmsg)

            cpyname, scripts_location = read_cfg(infile, check_python=False)
    else:
        cpyname, scripts_location = "", [""]

    cpyname, scripts_location = INPUT_DIALOG(
        title="XCPy Configuration",
        header="Only one directory path per line. The order for the folders\n"
               "given here is the order of priority in which they will be searched.\n"
               "Click on OK to write this configuration.",
        items=["CPython Executable", "CPython Scripts Location"],
        values=[cpyname, "\n".join(scripts_location)],
        comments=["", ""],
        types=["1", "5"],
        columns=40,
    )

    scripts_location = scripts_location.replace("\n", ",")

    if not cpyname or not scripts_location:
        MSG("Invalid configartion specified. Config file not written")
    else:
        config = SafeConfigParser()
        config.add_section("xcpy")
        config.set("xcpy", "cpython", cpyname)
        config.set("xcpy", "scripts_location", scripts_location)

        with open(outfile, "w") as f:
            config.write(f)
        MSG("Written Configuration file at: " + outfile)


def exists(filename, raise_error=False):
    """
    Checks whether a file exists either returns False or
    raises an exception

    """
    if os.path.exists(filename):
        return True
    elif raise_error:
        raise Exception("{} not found".format(filename))
    else:
        return False


def current_data():
    """
    Returns the current EXPNO and PROCNO open in Topspin,
    if executed when a data folder is open

    """
    cd = CURDATA()

    if cd is not None:
        current_dir = os.path.join(cd[3], cd[0])
        current_expno = cd[1]
        current_procno = cd[2]
        return [current_dir, current_expno, current_procno]

    else:
        MSG(
            """No data folder seems to be open!
            No arguments will be passed on. If this is intentional,
            you should run the command with the --no-args option,
            which will get rid of this message"""
        )
        return []


def get_scriptname():
    """
    Opens up a dialog box to get the script name

    """
    scriptname = INPUT_DIALOG(
        "Script",
        "Type the full path and name of the script to be run",
        ["CPython Script"],
        [""],
        [""],
        ["1"],
    )

    return scriptname


def run(cpython, script, pass_current_folder=True, use_shell=None, dry=None):
    """
    Runs a cpython script

    """
    if pass_current_folder is True:
        cd = current_data()
    else:
        cd = []

    if use_shell is None:
        if os.name == "nt":
            use_shell = True
        else:
            use_shell = False

    args = [cpython, script] + cd

    if dry:
        MSG("The following command will be executed: \n" + " ".join(args))
        process = None

    else:
        process = Popen(args, stdin=PIPE, stdout=PIPE, stderr=STDOUT, shell=use_shell)
        process.stdin.close()

    return process


def verify_completion(process):
    """
    Verify that the output is correct
    """

    if process is not None:
        errmsg = [line for line in iter(process.stdout.readline, "")]

        if not errmsg:
            MSG("Script executed to completion")
        else:
            MSG("".join(errmsg))

    else:
        return


def verify_python(command):
    """
    Verify that the command to be saved in the config file points
    to a valid python executable. This is a rudimentary check!

    """
    if not os.path.exists(command):
        raise OSError("The command {} does not seem to exist".format(command))

    command = command.split(os.sep)[-1]
    if command.lower().find("python") != 0:

        errmsg = """
            {} does not seem to be a valid python file.
            Please check the configuration file using 'xcpy --config',
            or change the congiguration file using 'xcpy --settings'
            This attempt will be aborted.
            """.format(
            command
        )

        raise OSError(errmsg)


def show_config(filename, printing=True):
    """
    Shows the configuration file if present

    """
    try:
        with open(filename, "r") as f:
            config = f.read()
            if printing:
                MSG(config)
    except FileNotFoundError:
        if printing:
            MSG(filename + " not found")
        config = None

    return config


def main():
    """
    The main function that gets called when xcpy is run

    """

    # check if the program is running via topspin
    # and get the path of topspin
    if check_jython():
        toppath = topspin_location()

    # path and name of the configuration file is hard-coded
    config_file = os.path.join(toppath, "exp", "stan", "nmr", "py", "user", "xcpy.cfg")

    # get arguments passed to xcpy
    # if no arguments are passed, tag on --help
    # so that the docstring shows up in the next step
    argv = sys.argv
    if len(argv) == 1:
        argv.append("--help")

    # Case: return docstring
    if argv[1] in ["-h", "--help"]:
        if len(argv) > 2:
            MSG("Opening Documentation. All other options ignored. Press OK")
        MSG(__doc__)

    # check if configuration exists
    # if it does not, alert the user and open up a dialog box to write it out
    elif not os.path.exists(config_file):
        MSG(
            """
            Configuration file does not exist.
            An input box will be opened next to write it.
            Alternately, it can be manually edited at {}.
            Press 'Close' to continue.
            """.format(
                config_file
            )
        )
        write_cfg(config_file)

    # if configuration settings are to be changed,
    # open up a dialog box to do so
    elif argv[1] in ["-s", "--settings"]:
        if len(argv) > 2:
            MSG("Opening configuration settings. Ignored all other options")
        write_cfg(config_file, config_file)

    # show configuration
    # prints out the contents of the configuration file
    elif argv[1] in ["-c", "--config"]:
        show_config(config_file)

    # if -h, -c, -s are not used, actual script run is required
    else:

        # STEP 1: Check for flags that need to be set

        # careful!, uses shell and only sanity check is
        # that the script being run is called 'python'
        # this is required for windows and by default off for *nix
        if "--use-shell" in argv:
            use_shell = True
        else:
            use_shell = False

        # if no arguments are to be passed to the script
        if "--no-args" in argv:
            pass_current_folder = False
        else:
            pass_current_folder = True


        # prints out the exact command that will be run using subprocess
        # does not actually run anything
        if "--dry-run" in argv or "-d" in argv:
            dry = True
        else:
            dry = False

        # STEP 2: READ in configuration file
        cpyname, folders = read_cfg(config_file)

        # STEP 3: See what script needs to be run
        # If a script at unknown location is to be used
        if "-n" in argv or "--name" in argv:
            scriptname = get_scriptname()
        else:
            # see if script is in one of the folders that
            # are given in the cfg file
            executed = False

            # search priority is top to bottom in the given list
            # if a file is found, it stops searching
            for folder in folders:
                scriptname = os.path.join(folder, argv[-1])

                # check for .py extension and append it if not given
                if not scriptname.endswith('.py'):
                    scriptname = scriptname + '.py'
                
                # run the script if it exists and then break from the for loop
                # and set the executed status to true
                if exists(scriptname):
                    process = run(
                        cpyname, scriptname, pass_current_folder, use_shell, dry
                    )
                    executed = True
                    break

            # executed should be false iff no script was found
            if not executed:
                raise Exception(
                    "The file {} was not found in the following folders:\n\n{}".format(
                        argv[-1] + ".py", "\n".join(folders)
                    )
                )

        # handle errors and print messages from cpython
        verify_completion(process)


if __name__ == "__main__":
    main()



"""
Functions for reading and writing spectral data to the glue format which
stores spectral data in the Hierarchical Data Format (HDF5).

glue files are HDF5 files with the spectral data stored in a dataset names
'spectrum' and any parameters stored in the dataset attributes.  At minimum
the parameter dictionary must contain a ndim key with the dimentionality of
the data and a dictionry for each axis numbered (0,1,2...) with the following
keys:

========    =====================================================
Key         Description
========    =====================================================
car         Carrier freqiency in Hz
complex     True for complex data, False for magnitude
encoding    type of encoding, 'states', 'tppi', 'direct', etc
freq        True for frequency domain data, False otherwise
label       Axis label, ('13C', etc)
obs         Observation frequency in MHz
size        Dimension size (R|I for direct dim, R+I for indirect)
sw          Spectral width in Hz
time        True for time domain data, False otherwise
========    =====================================================

"""

import numpy as np
import h5py
from . import fileiobase


# unit conversion functions
def make_uc(dic, data, dim=-1):
    """
    make a unit conversion object
    """
    if dim == -1:
        dim = data.ndim - 1  # last dimention

    size = dic[dim]["size"]
    cplex = dic[dim]["complex"]
    sw = dic[dim]["sw"]
    obs = dic[dim]["obs"]
    car = dic[dim]["car"]

    return fileiobase.unit_conversion(size, cplex, sw, obs, car)


# dictionary/data creation functions (null functions)
def create_data(data):
    """
    Create glue data array (return data)
    """
    return data


def guess_udic(dic, data):
    """
    Guess parameters of a universal dictionary from dic,data pair
    """
    return dic


def create_dic(udic):
    """
    Create a glue dictionary from a universal dictionary
    """
    return udic


# file reading/writing functions
def read(filename, dataset="spectrum"):
    """
    Read a glue file
    """
    dic, data = read_lowmem(filename, dataset)
    return dic, data[:]


def read_lowmem(filename, dataset="spectrum"):
    """
    Read a glue file using mimimal memory usage
    """
    f = h5py.File(filename, 'r')
    dic = get_dic(f, dataset)
    data = wrap_data(f[dataset])
    return dic, data


def write(filename, dic, data, dataset="spectrum", overwrite=False):
    """
    Write dic,data pair to a HDF5 file
    """
    # create the file
    f = fileiobase.open_towrite(filename, overwrite=overwrite)
    f.close()
    f = h5py.File(filename, 'w')

    # write the dictionary and data to the file
    f.create_dataset(dataset, data=data)
    put_dic(f, dic, dataset)

    f.close()
    return


# data wrapping functions
def wrap_data(data):
    """
    wrap h5py.highlevel.Dataset objects into a more numpy like object
    """
    if len(data.shape) == 1:
        return data
    elif len(data.shape) == 2:
        return glue_2d(data)
    elif len(data.shape) == 3:
        return glue_3d(data)
    else:
        return data


# dictionary get/put
def get_dic(f, dataset="spectrum"):
    """
    Get a dictionary from dataset in a HDF5 File
    """

    # select the data set
    dset = f[dataset]

    dic = {}
    # loop over the attributes
    for key, value in dset.attrs.items():

        if "_" in key:
            # we have an axis key
            axis, subkey = key.split("_", 1)
            axis = int(axis)
            if axis not in dic:
                dic[axis] = {}
            dic[axis][subkey] = value
        else:
            dic[key] = value

    return dic


def put_dic(f, dic, dataset="spectrum"):
    """
    Put a dictionary to the dataset in a HDF5 File
    """

    # select the data set
    dset = f[dataset]

    for key, value in dic.items():

        # axis dictionaries
        if type(key) == int and type(value) == dict:
            axis = key
            for axiskey, axisvalue in value.items():
                fullkey = str(axis) + "_" + axiskey
                dset.attrs[fullkey] = axisvalue

        # simple value
        else:
            dset.attrs[key] = value
    return


# glue_* objects
class glue_2d(fileiobase.data_nd):
    """
    glue_2d emulates numpy.ndarray objects without loading data into memory

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes functions create a new fid_2d object with the
      new axes ordering.
    * has ndim, shape, and dtype attributes.

    """

    def __init__(self, Dataset, order=["y", "x"]):
        """
        Create and set up object
        """
        # set up array attributes
        self.Dataset = Dataset
        self.lenX = int(Dataset.shape[1])
        self.lenY = int(Dataset.shape[0])
        a = [self.lenY, self.lenX]
        self.order = order
        self.shape = tuple([a[order.index(k)] for k in ["y", "x"]])
        self.ndim = 2
        self.dtype = Dataset.dtype

    def __fcopy__(self, order):
        """
        Create a copy
        """
        n = glue_3d(self.Dataset, order)
        return n

    def __fgetitem__(self, slices):
        """
        Return ndarray of selected values

        (sY,sX) is a well formated tuple of slices
        """
        sY, sY = slices
        return self.Dataset[sY, sX]


class glue_3d(fileiobase.data_nd):
    """
    glue_3d emulates numpy.ndarray objects without loading data into memory

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes functions create a new fid_2d object with the
      new axes ordering.
    * has ndim, shape, and dtype attributes.

    """

    def __init__(self, Dataset, order=["z", "y", "x"]):
        """
        Create and set up object
        """
        # set up array attributes
        self.Dataset = Dataset
        self.lenX = int(Dataset.shape[2])
        self.lenY = int(Dataset.shape[1])
        self.lenZ = int(Dataset.shape[0])
        a = [self.lenZ, self.lenY, self.lenX]
        self.order = order
        self.shape = tuple([a[order.index(k)] for k in ["z", "y", "x"]])
        self.ndim = 3
        self.dtype = Dataset.dtype

    def __fcopy__(self, order):
        """
        Create a copy
        """
        n = glue_3d(self.Dataset, order)
        return n

    def __fgetitem__(self, slices):
        """
        Return ndarray of selected values

        (sZ, sY, sX) is a well formated tuple of slices
        """
        sZ, sY, sX = slices
        return self.Dataset[sZ, sY, sX]



"""
Functions for reading and writing Sparky (.ucsf) files.
"""

from __future__ import print_function

__developer_info__ = """
Sparky file format information
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Information on the Sparky file format can be found online at:
http://www.cgl.ucsf.edu/home/sparky/manual/files.html
and in the Sparky source file ucsffile.cc.

"""

import os
import struct
import datetime
from warnings import warn
try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser

import numpy as np

from . import fileiobase


# unit conversion function
def make_uc(dic, data, dim=-1):
    """
    Create a unit conversion object.

    Parameters
    ----------
    dic : dict
        Dictionary of Sparky parameters.
    data : ndarray
        Array of NMR data.
    dim : int, optional
        Dimension number to create unit conversion object for.  Default is for
        last dimension.

    Returns
    -------
    uc : unit conversion object.
        Unit conversion object for given dimension.

    """
    if dim == -1:
        dim = data.ndim - 1  # last dimention

    wdic = dic["w" + str(int(1 + dim))]

    size = float(wdic["npoints"])
    cplx = False
    sw = wdic["spectral_width"]
    obs = wdic["spectrometer_freq"]
    car = wdic["xmtr_freq"] * obs

    return fileiobase.unit_conversion(size, cplx, sw, obs, car)


# dictionary/data creation
def create_data(data):
    """
    Create a Sparky data array (recast into float32 array)
    """
    return np.array(data, dtype="float32")


def guess_udic(dic, data):
    """
    Guess parameter of universal dictionary from dic,data pair.

    Parameters
    ----------
    dic : dict
        Dictionary of Sparky parameters.
    data : ndarray
        Array of NMR data.

    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameter.

    """
    # create an empty universal dictionary
    udic = fileiobase.create_blank_udic(data.ndim)

    # update default values
    for i in range(data.ndim):
        adic = dic["w" + str(i + 1)]
        udic[i]["size"] = data.shape[i]
        udic[i]["sw"] = adic['spectral_width']
        udic[i]["obs"] = adic['spectrometer_freq']
        udic[i]["car"] = adic['xmtr_freq'] * adic['spectrometer_freq']
        udic[i]["label"] = adic['nucleus']
        udic[i]["complex"] = False
        udic[i]["time"] = False
        udic[i]["freq"] = True

    return udic


def create_dic(udic, datetimeobj=datetime.datetime.now(), user='user'):
    """
    Create a Sparky parameter dictionary from universal dictionary.

    Parameters
    ----------
    udic : dict
        Universal dictionary of spectral parameters.
    datatimeobj : datetime object, optional
        Datetime to record in Sparky dictionary
    user : str, optional
        Username to record in Sparky dictionary. Default is 'user'

    Returns
    -------
    dic : dict
        Dictionary of Sparky parameters.

    """
    dic = dict()

    # determind shape of array
    shape = [udic[k]["size"] for k in range(udic["ndim"])]

    # populate the dictionary
    dic["ident"] = 'UCSF NMR'
    dic["naxis"] = udic["ndim"]
    dic["ncomponents"] = 1
    dic["encoding"] = 0
    dic["version"] = 2
    dic["owner"] = user
    dic["date"] = datetimeobj.ctime()
    dic["comment"] = ''
    dic["scratch"] = ''

    # calc a good tile shape
    tshape = calc_tshape(shape)

    # total number of tiles
    ntiles = 1
    for tlen, slen in zip(tshape, shape):
        ntiles *= np.ceil(float(slen) / tlen)

    # points in tile
    tpoints = np.array(tshape).prod()

    # data bytes
    dbytes = tpoints * ntiles * 4

    # total file size if data size plus leaders
    dic["seek_pos"] = int(dbytes + 180 + 128 * len(shape))

    # populate the dictionary with axis dictionaries
    for i, (tlen, dlen) in enumerate(zip(tshape, shape)):
        dic["w" + str(i + 1)] = create_axisdic(udic[i], tlen, dlen)
    return dic


def create_axisdic(adic, tlen, dlen):
    """
    Make an Sparky axis dictionary from a universal axis dictionary.

    Parameters
    ----------
    adic : dict
        Axis dictionary from a universal dictionary.
    tlen : int
        Tile length of axis.
    dlen : int
        Data length of axis.

    Returns
    -------
    sdic : dict
        Sparky axis dictionary

    """
    dic = dict()
    dic["nucleus"] = adic["label"]
    dic["spectral_shift"] = 0
    dic["npoints"] = int(dlen)
    dic["size"] = int(dlen)
    dic["bsize"] = int(tlen)
    dic["spectrometer_freq"] = float(adic["obs"])
    dic["spectral_width"] = float(adic["sw"])
    dic["xmtr_freq"] = float(adic["car"]) / dic["spectrometer_freq"]
    dic["zero_order"] = 0.0
    dic["first_order"] = 0.0
    dic["first_pt_scale"] = 0.0
    dic["extended"] = b'\x80'  # transform bit set
    return dic


def datetime2dic(datetimeobj, dic):
    """
    Add datetime object to dictionary
    """
    dic["date"] = datetimeobj.ctime()
    return dic


def dic2datetime(dic):
    """
    Create a datetime object from a Sparky dictionary
    """
    return datetime.datetime.strptime(dic["date"], "%a %b %d %H:%M:%S %Y")


def calc_tshape(shape, kbyte_max=128):
    """
    Calculate a tile shape from data shape.


    Parameters
    ----------
    shape : tuple
        Shape of NMR data (data.shape).
    kbyte_max : float or int
        Maximum tile size in Kilobytes.

    Returns
    -------
    tshape : tuple
        Shape of tile.

    """
    # Algorithm divides each dimention by 2 until under kbyte_max tile size.
    s = np.array(shape, dtype="int")
    i = 0
    while (s.prod() * 4. / 1024. > kbyte_max):
        s[i] = np.floor(s[i] / 2.)
        i = i + 1
        if i == len(s):
            i = 0
    return tuple(s)


# global read/write functions
def read(filename):
    """
    Read a Sparky file.

    Parameters
    ----------
    filename : str
        Filename of Sparky file to read.

    Returns
    -------
    dic : dict
        Dictionary of Sparky parameters.
    data : ndarray
        Array of NMR data.

    See Also
    --------
    read_lowmem : Sparky file reading with minimal memory usage.
    write : Write a Sparky file.

    """
    # open the file
    f = open(filename, 'rb')

    # determind the dimentionality
    n = fileheader2dic(get_fileheader(f))["naxis"]
    f.close()

    if n == 2:
        return read_2D(filename)
    if n == 3:
        return read_3D(filename)

    raise ValueError("unknown dimentionality: %s" % n)


def read_lowmem(filename):
    """
    Read a Sparky file using minimal memory.

    Parameters
    ----------
    filename : str
        Filename of Sparky file to read.

    Returns
    -------
    dic : dict
        Dictionary of Sparky parameters.
    data : array_like
        Low memory object which can access NMR data on demand.

    See Also
    --------
    read : Read a Sparky file.
    write_lowmem : Write a Sparky file using mimimal memory.

    """
    # open the file
    f = open(filename, 'rb')

    # determind the dimentionality
    n = fileheader2dic(get_fileheader(f))["naxis"]
    f.close()

    if n == 2:
        return read_lowmem_2D(filename)
    if n == 3:
        return read_lowmem_3D(filename)

    raise ValueError("unknown dimentionality: %s" % n)


def write(filename, dic, data, overwrite=False):
    """
    Write a Sparky file.

    Parameters
    ----------
    filename : str
        Filename of Sparky file to write to.
    dic : dict
        Dictionary of Sparky parameters.
    data : array_like
        Array of NMR data.
    overwrite : bool, optional
        Set True to overwrite files, False will raise a Warning if the file
        exists.

    See Also
    --------
    write_lowmem : Write a Sparky file using minimal amounts of memory.
    read : Read a Sparky file.

    """
    n = dic["naxis"]

    if n == 2:
        return write_2D(filename, dic, data, overwrite=overwrite)
    if n == 3:
        return write_3D(filename, dic, data, overwrite=overwrite)

    raise ValueError("unknown dimentionality: %s" % n)


def write_lowmem(filename, dic, data, overwrite=False):
    """
    Write a Sparky using minimum amounts of memory (tile by tile)

    Parameters
    ----------
    filename : str
        Filename of Sparky file to write to.
    dic : dict
        Dictionary of Sparky parameters.
    data : array_like.
        Array of NMR data.
    overwrite : bool, optional
        Set True to overwrite files, False will raise a Warning if the file
        exists.

    See Also
    --------
    write : Write a Sparky file.
    read_lowmem : Read a Sparky file using mimimal amounts of memory.

    """
    # write also writes tile by tile...
    return write(filename, dic, data, overwrite)


# dimension specific reading/writing functions
def read_2D(filename):
    """
    Read a 2D sparky file. See :py:func:`read` for documentation.
    """
    seek_pos = os.stat(filename).st_size
    with open(filename, 'rb') as f:

        # read the file header
        dic = fileheader2dic(get_fileheader(f))

        # check for file size mismatch
        if seek_pos != dic["seek_pos"]:
            warn('Bad file size in header %s vs %s' %
                 (seek_pos, dic['seek_pos']))

        # read the axis headers...
        for i in range(dic['naxis']):
            dic["w" + str(i + 1)] = axisheader2dic(get_axisheader(f))

        # read the data and untile
        lenY = dic["w1"]["npoints"]
        lenX = dic["w2"]["npoints"]
        lentY = dic["w1"]["bsize"]
        lentX = dic["w2"]["bsize"]
        data = get_data(f)
        data = untile_data2D(data, (lentY, lentX), (lenY, lenX))

        return dic, data


def write_2D(filename, dic, data, overwrite=False):
    """
    Write a 2D Sparky file. See :py:func:`write` for documentation.
    """
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite)

    # write the file header
    put_fileheader(f, dic2fileheader(dic))

    # write the axis headers
    put_axisheader(f, dic2axisheader(dic["w1"]))
    put_axisheader(f, dic2axisheader(dic["w2"]))

    lentX = dic["w2"]["bsize"]
    lentY = dic["w1"]["bsize"]
    t_tup = (lentY, lentX)

    ttX = int(np.ceil(data.shape[1] / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(data.shape[0] / float(lentY)))  # total tiles in Y dim
    tt = ttX * ttY

    for i in range(int(tt)):
        put_data(f, find_tilen_2d(data, i, (t_tup)))

    f.close()
    return


def read_3D(filename):
    """
    Read a 3D Sparky file. See :py:func:`read` for documentation.
    """
    seek_pos = os.stat(filename).st_size
    with open(filename, 'rb') as f:

        # read the file header
        dic = fileheader2dic(get_fileheader(f))

        # check for file size mismatch
        if seek_pos != dic["seek_pos"]:
            warn('Bad file size in header %s vs %s' %
                 (seek_pos, dic['seek_pos']))

        # read the axis headers...
        for i in range(dic['naxis']):
            dic["w" + str(i + 1)] = axisheader2dic(get_axisheader(f))

        # read the data and untile
        lenZ = dic["w1"]["npoints"]
        lenY = dic["w2"]["npoints"]
        lenX = dic["w3"]["npoints"]
        lentZ = dic["w1"]["bsize"]
        lentY = dic["w2"]["bsize"]
        lentX = dic["w3"]["bsize"]
        data = get_data(f)
        data = untile_data3D(data, (lentZ, lentY, lentX), (lenZ, lenY, lenX))

        return dic, data


def write_3D(filename, dic, data, overwrite=False):
    """
    Write a 3D Sparky file. See :py:func:`write` for documentation.
    """
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite)

    # write the file header
    put_fileheader(f, dic2fileheader(dic))

    # write the axis headers
    put_axisheader(f, dic2axisheader(dic["w1"]))
    put_axisheader(f, dic2axisheader(dic["w2"]))
    put_axisheader(f, dic2axisheader(dic["w3"]))

    lentX = dic["w3"]["bsize"]
    lentY = dic["w2"]["bsize"]
    lentZ = dic["w1"]["bsize"]

    t_tup = (lentZ, lentY, lentX)

    ttX = int(np.ceil(data.shape[2] / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(data.shape[1] / float(lentY)))  # total tiles in Y dim
    ttZ = int(np.ceil(data.shape[0] / float(lentZ)))  # total tiles in Z dim

    tt = ttX * ttY * ttZ

    for i in range(int(tt)):
        put_data(f, find_tilen_3d(data, i, (t_tup)))
    f.close()
    return


# read_lowmem functions
def read_lowmem_2D(filename):
    """
    Read a 2D Sparky file using minimal memory. See :py:func:`read_lowmem`.
    """
    seek_pos = os.stat(filename).st_size

    # create the sparky_2d file
    data = sparky_2d(filename)
    dic = dict(data.dic)

    # check for file size mismatch
    if seek_pos != dic["seek_pos"]:
        warn('Bad file size in header %s vs %s' % (seek_pos, dic['seek_pos']))
    return dic, data


def read_lowmem_3D(filename):
    """
    Read a 3D sparky file using minimal memory. See :py:func:`read_lowmem`.
    """
    seek_pos = os.stat(filename).st_size

    # create the sparky_3d file
    data = sparky_3d(filename)
    dic = dict(data.dic)

    # check for file size mismatch
    if seek_pos != dic["seek_pos"]:
        warn('Bad file size in header %s vs %s' % (seek_pos, dic['seek_pos']))

    return dic, data


class SparkySaveParser(HTMLParser):
    """
    A parser for Sparky .save files. The file structure is similar to
    simple HTML files, except the use of <end tag> instead of </tag>.
    The following structure is assumed:

    <sparky save file>
    <version ...>
    <user>
    ...
    <end user>
    <spectrum>
    ...
        <view>
        ...
            <params>
            ...
            <end params>
            <params>
            ...
            <end params>
        <end view>
        <view>
        ...
            <params>
            ...
            <end params>
            <params>
            ...
            <end params>
        <end view>
        <ornament>
        ...
        <end ornament>
    <end spectrum>

    TODO: some .save files do not have this exact structure
    They need to be treated differently

    """

    # main dictionaries to parse data into
    user, spectrum, view, ornament = {}, {}, {}, {}

    # tracker if there are multiple views
    viewnum = -1

    curtag, curdict = None, None

    def _parse_info(self, string, dtype=None):
        """
        Reads a list of strings into a dictionary, with the first item of
        the list as the key and the remaining list as the value. In addition,
        it parses all values in the list to do the following: (i) convert the
        values to float wherever possible and (ii) if the list has a single
        item, upack and return that item alone as the value

        """

        dic = {}
        for s in string:
            i = s.split()
            try:
                if dtype is None:
                    key = i[0].replace(".", "_").replace("-", "_")
                    value = i[1:]

                if dtype == "user":
                    key = i[0] + "_" + i[1]
                    value = i[2:]

                parsed_value = []
                for v in value:
                    try:
                        if key == "id":
                            parsed_value.append(int(v))
                        else:
                            parsed_value.append(float(v))
                    except ValueError:
                        parsed_value.append(v)

                if len(value) == 1:
                    dic[key] = parsed_value[0]

                else:
                    dic[key] = parsed_value

            except IndexError:
                pass

        return dic

    def _parse_peak(self, peak):
        """
        Parses a single peak into a dictionary, the input being a list
        that corresponds to a single peak in a sparky save file. In addition,
        it parses all values in the list to do the following: (i) convert the
        values to float wherever possible and (ii) if the list has a single
        item, upack and return that item alone as the value. Currently assumes
        the following structure for a single peak:

        type peak
        ...
        [
        type label
        ...
        ]

        """

        l = [i for i, word in enumerate(peak) if word in ["[", "]"]]
        p = peak[:l[0]]
        l = peak[l[0]+1:l[1]]

        d = {}
        for i in p:
            j = i.split()

            info = []
            for k in j:
                try:
                    info.append(float(k))
                except ValueError:
                    info.append(k)

            if len(info) == 2:
                d[info[0]] = info[1]
            else:
                d[info[0]] = info[1:]

        for i in l:
            j = i.split()
            if j[0] not in d.keys():
                d[j[0]] = j[1:]
        del d["type"]

        for i, k in enumerate(d["xy"]):
            k = k.split(",")
            d["xy"][i] = [float(j) for j in k]

        return d

    def _parse_ornaments(self, data):
        """
        Parses a string containing all ornaments into a dictionary. This
        is for all the data inside the <ornament> tag. The key for each
        peak item is given by the peak ID, which should be unique for each
        peak. The following structure is assumed:

        type peak # peak 1
        ...
        type peak # peak 2
        ...
        type peak # peak 3
        ...

        """

        data = data.split("\n")
        p = [i for i, word in enumerate(data) if word == "type peak"]
        peaklist = [data[p[i]: p[i+1]] for i in range(len(p)-1)]

        dic = {}
        for peak in peaklist:
            d = self._parse_peak(peak)
            dic[int(d["id"])] = d

        return dic


    def handle_starttag(self, tag, attrs):

        if tag in ["sparky", "version"]:
            self.curdict = self.spectrum
            self.spectrum[tag] = attrs[0][0]

        elif tag == "user":
            self.curdict = self.user

        elif tag == "spectrum":
            self.curdict = self.spectrum

        elif tag == "view":
            self.viewnum += 1
            self.view[self.viewnum] = {}
            self.curdict = self.view[self.viewnum]

        elif tag == "ornament":
            self.curdict = self.ornament

        else:
            # params tag for each view
            self.curtag = tag


    def handle_endtag(self, tag):
        self.curtag = None


    def handle_data(self, data):

        # ignore blank lines
        if len(data.strip()) == 0:
            pass

        elif self.curtag not in self.curdict.keys():

            # all the files that are read in a split at the newline character
            if (self.curtag is None) and (self.curdict == self.spectrum):
                dic = self._parse_info(data.split("\n"),)
                for k, v in dic.items():
                    self.curdict[k] = v

            elif (self.curtag is None) and (self.curdict == self.user):
                dic = self._parse_info(data.split("\n"), "user")
                for k, v in dic.items():
                    self.curdict[k] = v

            elif (self.curtag is None) and (self.curdict == self.view[self.viewnum]):
                dic = self._parse_info(data.split("\n"),)
                for k, v in dic.items():
                    self.curdict[k] = v

            elif (self.curtag is None) and (self.curdict == self.ornament):
                self.ornament = self._parse_ornaments(data)

            else:
                # for a param tag inside a view
                dic = self._parse_info(data.split("\n"),)
                self.curdict[self.curtag] = [dic]

        else:
            # this is only executed for multiple params tags in a view
            dic = self._parse_info(data.split("\n"),)
            self.curdict[self.curtag].append(dic)


def read_savefile(savefile, spectrum_file=None):
    """
    Reads in a Sparky .save file and the corresponding spectrum (.ucsf)
    file. In addition to the usual dictionary contents that come with
    a .ucsf file, these additinal dictionary keys are created with the content
    from .save file: "spectrum", "view", "user" and "ornament". The together
    contain all edits and annotations. By default, it tries to read in
    the spectrum file given in the .save file (but this fails many times due
    to relative paths in .save file)

    Parameters
    ----------
    savefile : str
        Filename of Sparky .save file.
    spectrum_file : str
        Filename of Sparky .ucsf file.


    Returns
    -------
    dic : dict
        Dictionary of Sparky .ucsf and .save parameters.
    data : ndarray
        Array of NMR data.

    """

    with open(savefile, "r") as f:
        savefile = f.read().replace("<end ", r"</")

    parser = SparkySaveParser()
    parser.feed(savefile)
    parser.close()

    dic = {
        "user": parser.user,
        "spectrum": parser.spectrum,
        "view": parser.view,
        "ornament": parser.ornament,
    }

    try:
        if spectrum_file is None:
            d, data = read(dic["spectrum"]["pathname"])
        else:
            d, data = read(spectrum_file)

    except:
        warn("Cannot load spectrum")
        d, data = {}, None

    dic.update(d)

    return dic, data


# sparky_ low memory objects
class sparky_2d(fileiobase.data_nd):
    """
    Emulates a ndarray object without loading data into memory for low memory
    reading of 2D Sparky files.

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes methods create a new objects with correct axes
      ordering.
    * has ndim, shape, and dtype attributes.

    Parameters
    ----------
    filename : str
        Filename of 2D Sparky file.
    order : tuple, optional
        Order of axes against file.  None is equivelent to (0, 1).

    """

    def __init__(self, filename, order=None):
        """
        Create and set up object
        """
        # open the file
        self.filename = filename
        f = open(filename, 'rb')

        # read the fileheader
        self.dic = fileheader2dic(get_fileheader(f))

        if self.dic["naxis"] != 2:
            raise Exception("file is not a 2D Sparky file")

        # read in the axisheaders
        self.dic["w1"] = axisheader2dic(get_axisheader(f))
        self.dic["w2"] = axisheader2dic(get_axisheader(f))
        f.close()

        # sizes
        self.lenY = self.dic["w1"]["npoints"]
        self.lenX = self.dic["w2"]["npoints"]

        # tile sizes
        self.lentY = self.dic["w1"]["bsize"]
        self.lentX = self.dic["w2"]["bsize"]

        # check order
        if order is None:
            order = (0, 1)

        # finalize
        self.dtype = np.dtype("float32")
        self.order = order
        self.fshape = (self.lenY, self.lenX)
        self.__setdimandshape__()

    def __fcopy__(self, order):
        """
        Create a copy
        """
        n = sparky_2d(self.filename, order)
        return n

    def __fgetitem__(self, slices):
        """
        Returns ndarray of selected values.

        (sY, sX) is a well formatted tuple of slices
        """
        sY, sX = slices
        f = open(self.filename, 'rb')

        # print(sY,sX)
        gY = range(self.lenY)[sY]  # list of values to take in Y
        gX = range(self.lenX)[sX]  # list of values to take in X

        # tiles to get in each dim to read
        # Y tile to read
        gtY = set([int(np.floor(i / self.lentY)) for i in gY])
        # X tile to read
        gtX = set([int(np.floor(i / self.lentX)) for i in gX])

        # create a empty output directory
        out = np.empty((len(gY), len(gX)), dtype=self.dtype)

        for iY in gtY:      # loop over Y tiles to get
            for iX in gtX:  # loop over X tiles to get

                # get the tile and reshape it
                ntile = int(iY * np.ceil(self.lenX / self.lentX) + iX)
                tile = get_tilen(f, ntile, (self.lentX, self.lentY))
                tile = tile.reshape(self.lentY, self.lentX)

                # tile minimum and max values for each dim
                minX = iX * self.lentX
                maxX = (iX + 1) * self.lentX

                minY = iY * self.lentY
                maxY = (iY + 1) * self.lentY

                # determind what elements are needed from this tile
                XinX = [i for i in gX if maxX > i >= minX]  # values in gX
                XinT = [i - minX for i in XinX]  # tile index values
                XinO = [gX.index(i) for i in XinX]  # output indexes

                YinY = [i for i in gY if maxY > i >= minY]  # values in gX
                YinT = [i - minY for i in YinY]  # tile index values
                YinO = [gY.index(i) for i in YinY]  # output indexes

                # take elements from the tile
                ctile = tile.take(XinT, axis=1).take(YinT, axis=0)

                # DEBUGGING info
                # print("-------------------------------")
                # print("iX:",iX,"iY:",iY,"ntile:",ntile)
                # print("tile.shape",tile.shape)
                # print("minX:",minX,"maxX",maxX)
                # print("minY:",minY,"maxY",maxY)
                # print("XinX",XinX)
                # print("XinT",XinT)
                # print("XinO",XinO)
                # print("YinY",YinY)
                # print("YinT",YinT)
                # print("YinO",YinO)

                # put the cut tile to the out array (uses some fancy indexing)
                out[np.ix_(YinO, XinO)] = ctile

        f.close()
        return out


class sparky_3d(fileiobase.data_nd):
    """
    Emulates a ndarray object without loading data into memory for low memory
    read of 3D Sparky files.

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes methods create a new objects with correct axes
      ordering.
    * has ndim, shape, and dtype attributes.

    Parameters
    ----------
    filename : str
        Filename of 3D Sparky file.
    order : tuple
        Ordering of axes against file. None is equilent to (0, 1, 2)

    """
    def __init__(self, filename, order=None):
        """
        Create and set up object
        """
        # open the file
        self.filename = filename
        f = open(filename, 'rb')

        # read the fileheader
        self.dic = fileheader2dic(get_fileheader(f))

        if self.dic["naxis"] != 3:
            raise Exception("file not 3D Sparky file")

        # read in the axisheaders
        self.dic["w1"] = axisheader2dic(get_axisheader(f))
        self.dic["w2"] = axisheader2dic(get_axisheader(f))
        self.dic["w3"] = axisheader2dic(get_axisheader(f))
        f.close()

        # sizes
        self.lenZ = self.dic["w1"]["npoints"]
        self.lenY = self.dic["w2"]["npoints"]
        self.lenX = self.dic["w3"]["npoints"]

        # tile sizes
        self.lentZ = self.dic["w1"]["bsize"]
        self.lentY = self.dic["w2"]["bsize"]
        self.lentX = self.dic["w3"]["bsize"]

        # check order
        if order is None:
            order = (0, 1, 2)

        # finalize
        self.dtype = np.dtype("float32")
        self.order = order
        self.fshape = (self.lenZ, self.lenY, self.lenX)
        self.__setdimandshape__()

    def __fcopy__(self, order):
        """
        Create a copy
        """
        n = sparky_3d(self.filename, order)
        return n

    def __fgetitem__(self, slices):
        """
        Returns ndarray of selected values.

        (sZ, sY, sX) is a well formateed tuple of slices
        """
        sZ, sY, sX = slices
        f = open(self.filename, 'rb')

        gZ = range(self.lenZ)[sZ]  # list of values to take in Z
        gY = range(self.lenY)[sY]  # list of values to take in Y
        gX = range(self.lenX)[sX]  # list of values to take in X

        # tiles to get in each dim to read
        # Z tiles
        gtZ = set([int(np.floor(float(i) / self.lentZ)) for i in gZ])
        # Y tiles
        gtY = set([int(np.floor(float(i) / self.lentY)) for i in gY])
        # X tiles
        gtX = set([int(np.floor(float(i) / self.lentX)) for i in gX])

        # total tiles in each dim
        ttX = int(np.ceil(self.lenX / float(self.lentX)))  # total tiles in X
        ttY = int(np.ceil(self.lenY / float(self.lentY)))  # total tiles in Y
        ttZ = int(np.ceil(self.lenZ / float(self.lentZ)))  # total tiles in Z

        tile_tup = (self.lentZ, self.lentY, self.lentX)

        # create a empty output array
        out = np.empty((len(gZ), len(gY), len(gX)), dtype=self.dtype)

        for iZ in gtZ:          # loop over Z tiles to get
            for iY in gtY:      # loop over Y tiles to get
                for iX in gtX:  # loop over X tiles to get

                    # get the tile and reshape it
                    ntile = iZ * ttX * ttY + iY * ttX + iX
                    tile = get_tilen(f, ntile, tile_tup)
                    tile = tile.reshape(tile_tup)

                    # tile minimum and max values for each dim
                    minX = iX * self.lentX
                    maxX = (iX + 1) * self.lentX

                    minY = iY * self.lentY
                    maxY = (iY + 1) * self.lentY

                    minZ = iZ * self.lentZ
                    maxZ = (iZ + 1) * self.lentZ

                    # determind what elements are needed from this tile
                    XinX = [i for i in gX if maxX > i >= minX]  # values in gX
                    XinT = [i - minX for i in XinX]  # tile index values
                    XinO = [gX.index(i) for i in XinX]  # output indexes

                    YinY = [i for i in gY if maxY > i >= minY]  # values in gX
                    YinT = [i - minY for i in YinY]  # tile index values
                    YinO = [gY.index(i) for i in YinY]  # output indexes

                    ZinZ = [i for i in gZ if maxZ > i >= minZ]  # values in gX
                    ZinT = [i - minZ for i in ZinZ]  # tile index values
                    ZinO = [gZ.index(i) for i in ZinZ]  # output indexes

                    # take elements from the tile
                    ctile = tile.take(XinT, axis=2).take(YinT, axis=1)
                    ctile = ctile.take(ZinT, axis=0)

                    # DEBUGGING info
                    # print("-------------------------------")
                    # print("iX:",iX,"iY:",iY,"iZ:",iZ,"ntile:",ntile)
                    # print("ttX:",ttX,"ttY:",ttY,"ttZ",ttZ)
                    # print("tile.shape",tile.shape)
                    # print("minX:",minX,"maxX",maxX)
                    # print("minY:",minY,"maxY",maxY)
                    # print("minZ:",minZ,"maxZ",maxZ)
                    # print("XinX",XinX)
                    # print("XinT",XinT)
                    # print("XinO",XinO)
                    # print("YinY",YinY)
                    # print("YinT",YinT)
                    # print("YinO",YinO)
                    # print("ZinZ",ZinZ)
                    # print("ZinT",ZinT)
                    # print("ZinO",ZinO)

                    # put the cut tile to the out array
                    out[np.ix_(ZinO, YinO, XinO)] = ctile
        f.close()
        return out


# tile and data get/put functions
def get_tilen(f, n_tile, tw_tuple):
    """
    Read a tile from a Sparky file object.

    Parameters
    ----------
    f : file object
        Open file object pointing to a Sparky file.
    n_tile : int
        Tile number to read
    tw_tuple : tuple of ints
        Tile size

    Returns
    -------
    tile : ndarray
        Tile of NMR data. Data is returned as a 1D array.

    Notes
    -----
    Current file position is loss. In can be stored before calling if the
    position is later needed.

    """
    # determind the size of the tile in bytes
    tsize = 4
    for i in tw_tuple:
        tsize = tsize * i

    # seek to the beginning of the tile
    f.seek(int(180 + 128 * len(tw_tuple) + n_tile * tsize))
    return np.frombuffer(f.read(tsize), dtype='>f4')


def get_tile(f, num_points):
    """
    Read the next tile from a Sparky file object.

    Parameters
    ----------
    f : file object
        Open file object pointing to a Sparky file.
    num_points : int
        Number of points in the tile.

    Returns
    -------
    tile : ndarray
        Tile of NMR data. Data is returned as a 1D array.

    """
    bsize = num_points * 4        # size in bytes
    return np.frombuffer(f.read(bsize), dtype='>f4')


def put_tile(f, tile):
    """
    Put a tile to a Sparky file object.

    Parameters
    ----------
    f : file object
        Open file object pointing to a Sparky file, to be written to.
    tile : ndarray
        Tile of NMR data to be written.

    """
    f.write(tile.astype('>f4').tostring())
    return


def get_data(f):
    """
    Read all data from sparky file object.
    """
    return np.frombuffer(f.read(), dtype='>f4')


def put_data(f, data):
    """
    Put data to a Sparky file object.

    This function does not untile data. This should be done before calling
    this function

    """
    f.write(data.astype('>f4').tostring())
    return


# tiling/untiling functions
def find_tilen_2d(data, ntile, tile_size):
    """
    Return a tile from a 2D NMR data set.

    Parameters
    ----------
    data : 2D ndarray
        NMR data, untiled/standard format.
    ntile : int
        Tile number to extract.
    (lentY, lentX) : tuple of ints
        Tile size (w1, w2).

    Returns
    -------
    tile : 1D ndarray
        Tile of NMR data, returned as 1D array.

    Notes
    -----
    Edge tiles are zero filled to the indicated tile size.

    """
    lentY, lentX = tile_size
    ttX = int(np.ceil(data.shape[1] / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(data.shape[0] / float(lentY)))  # total tiles in Y dim

    # tile number in each dim
    Xt = ntile % ttX
    Yt = int(np.floor(ntile / ttX))

    # dimention limits
    Xmin = int(Xt * lentX)
    Xmax = int((Xt + 1) * lentX)

    Ymin = int(Yt * lentY)
    Ymax = int((Yt + 1) * lentY)

    tile = data[Ymin:Ymax, Xmin:Xmax]

    # some edge tiles might need zero filling
    # see if this is the case
    if tile.shape == (lentY, lentX):    # well sized tile
        return tile.flatten()
    else:
        new_tile = np.zeros((lentY, lentX), dtype="float32")
        new_tile[:tile.shape[0], :tile.shape[1]] = tile
        return new_tile.flatten()


def tile_data2d(data, tile_size):
    """
    Tile 2D data into a 1D array.

    Parameters
    ----------
    data : 2D ndarray
        NMR data, untiled/standard format.
    (lentY, lentX) : tuple of ints
        Tile size.

    Returns
    -------
    tdata : 1D ndarray
        Tiled/Sparky formatted NMR data, returned as 1D array.

    """
    lentY, lentX = tile_size
    # determind the number of tiles in data
    ttX = int(np.ceil(data.shape[1] / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(data.shape[0] / float(lentY)))  # total tiles in Y dim
    tt = ttX * ttY  # total number of tiles

    # calc some basic parameter
    tsize = lentX * lentY   # number of points in one tile
    t_tup = (lentY, lentX)  # tile size tuple

    # create an empty array to store file data
    out = np.empty((tt * tsize), dtype="float32")

    for i in range(int(tt)):
        out[i * tsize:(i + 1) * tsize] = find_tilen_2d(data, i, t_tup)

    return out


def untile_data2D(data, tile_size, data_size):
    """
    Rearrange 2D Tiled/Sparky formatted data into standard format.

    Parameters
    ----------
    data : 1D ndarray
        Tiled/Sparky formatted 2D NMR data.
    (lentY, lenX) : tuple of ints
        Size of tile.
    (lenY, lenX) : tuple of ints
        Size of NMR data.

    Returns
    -------
    sdata : 2D ndarray
        NMR data, untiled/standard format.

    """
    lentY, lentX = tile_size
    lenY, lenX = data_size
    # determind the number of tiles in data
    ttX = int(np.ceil(lenX / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(lenY / float(lentY)))  # total tiles in Y dim
    tt = ttX * ttY

    # calc some basic parameter
    tsize = lentX * lentY  # number of points in one tile
    t_tup = (lentY, lentX)  # tile size tuple

    # create an empty array to store file data
    out = np.empty((ttY * lentY, ttX * lentX), dtype="float32")

    for iY in range(int(ttY)):
        for iX in range(int(ttX)):
            minX = iX * lentX
            maxX = (iX + 1) * lentX

            minY = iY * lentY
            maxY = (iY + 1) * lentY

            ntile = iY * ttX + iX
            minT = ntile * tsize
            maxT = (ntile + 1) * tsize

            # DEBUG
            # print("ntile",ntile)
            # print("minX",minX,"maxX",maxX)
            # print("minY",minY,"maxY",maxY)
            # print("minT",minT,"maxT",maxT)

            # print(out[minY:maxY,minX:maxX].shape)
            # print(data[minT:maxT].reshape(t_tup).shape)

            out[minY:maxY, minX:maxX] = data[minT:maxT].reshape(t_tup)

    return out[:lenY, :lenX]


def find_tilen_3d(data, ntile, tile_size):
    """
    Return a single tile from a 3D NMR data set.

    Parameters
    ----------
    data : 3D ndarray
        NMR data, untiled/standard format.
    ntile : int
        Tile number to extract.
    (lentZ, lentY, lentX) : tuple of ints
        Tile size (w1, w2, w3).

    Returns
    -------
    tile : 1D ndarray
        Tile of NMR data, returned as 1D array.

    Notes
    -----
    Edge tiles are zero filled to the indicated tile size.

    """
    lentZ, lentY, lentX = tile_size
    ttX = int(np.ceil(data.shape[2] / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(data.shape[1] / float(lentY)))  # total tiles in Y dim
    ttZ = int(np.ceil(data.shape[0] / float(lentZ)))  # total tiles in Z dim

    # tile number in each dim
    Xt = ntile % ttX
    Yt = int(np.floor(ntile / ttX)) % ttY
    Zt = int(np.floor(ntile / (ttX * ttY)))

    # dimention limits
    Xmin = int(Xt * lentX)
    Xmax = int((Xt + 1) * lentX)

    Ymin = int(Yt * lentY)
    Ymax = int((Yt + 1) * lentY)

    Zmin = int(Zt * lentZ)
    Zmax = int((Zt + 1) * lentZ)

    tile = data[Zmin:Zmax, Ymin:Ymax, Xmin:Xmax]

    # some edge tiles might need zero filling
    # see if this is the case
    if tile.shape == (lentZ, lentY, lentX):  # well sized tile
        return tile.flatten()
    else:
        new_tile = np.zeros((lentZ, lentY, lentX), dtype="float32")
        new_tile[:tile.shape[0], :tile.shape[1], :tile.shape[2]] = tile
        return new_tile.flatten()


def tile_data3d(data, tile_size):
    """
    Tile 3D data into a 1D numpy array

    Parameters
    ----------
    data : 3D ndarray
        NMR data, untiled/standard format.
    (lentZ, lentY, lentX) : tuple of ints
        Tile size (w1, w2, w3).

    Returns
    -------
    tile : 1D ndarray
        Tiled/Sparky formatted NMR data, returned as 1D array.

    """
    lentZ, lentY, lentX = tile_size
    # determind the number of tiles in data
    ttX = int(np.ceil(data.shape[2] / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(data.shape[1] / float(lentY)))  # total tiles in Y dim
    ttZ = int(np.ceil(data.shape[0] / float(lentZ)))  # total tiles in Z dim

    tt = ttX * ttY * ttZ    # total number of tiles

    # calc some basic parameter
    tsize = lentX * lentY * lentZ  # number of points in one tile
    t_tup = (lentZ, lentY, lentX)  # tile size tuple

    # create an empty array to store file data
    out = np.empty((tt * tsize), dtype="float32")

    for i in range(int(tt)):
        out[i * tsize:(i + 1) * tsize] = find_tilen_3d(data, i, t_tup)
    return out


def untile_data3D(data, tile_size, data_size):
    """
    Rearrange 3D tiled/Sparky formatted data into standard format.

    Parameters
    ----------
    data : 1D ndarray
        Tiled/Sparky formatted 2D NMR data.
    (lentZ, lentY, lentX) : tuple of ints
        Size of tile
    (lenZ, lenY, lenX) : tuple of ints
        Size of NMR data.

    Returns
    -------
    sdata : 3D ndarray
        NMR data, untiled/standard format.

    """
    lentZ, lentY, lentX = tile_size
    lenZ, lenY, lenX = data_size

    # determind the number of tiles in data
    ttX = int(np.ceil(lenX / float(lentX)))  # total tiles in X dim
    ttY = int(np.ceil(lenY / float(lentY)))  # total tiles in Y dim
    ttZ = int(np.ceil(lenZ / float(lentZ)))  # total tiles in Z dim
    tt = ttX * ttY * ttZ

    # calc some basic parameter
    tsize = lentX * lentY * lentZ  # number of points in one tile
    t_tup = (lentZ, lentY, lentX)  # tile size tuple

    # create an empty array to store file data
    out = np.empty((ttZ * lentZ, ttY * lentY, ttX * lentX), dtype="float32")

    for iZ in range(int(ttZ)):
        for iY in range(int(ttY)):
            for iX in range(int(ttX)):

                minX = iX * lentX
                maxX = (iX + 1) * lentX

                minY = iY * lentY
                maxY = (iY + 1) * lentY

                minZ = iZ * lentZ
                maxZ = (iZ + 1) * lentZ

                ntile = iZ * ttX * ttY + iY * ttX + iX
                minT = ntile * tsize
                maxT = (ntile + 1) * tsize

                out[minZ:maxZ, minY:maxY, minX:maxX] =  \
                    data[minT:maxT].reshape(t_tup)

    return out[:lenZ, :lenY, :lenX]


# fileheader functions
def get_fileheader(f):
    """
    Get fileheader from file and return a list.

    Reads the 180 byte file header of a Sparky file

    """
    # file header as descriped in ucsffile.cc of sparky source
    # header is packed as follows:
    # ident(10s),naxis(c),ncomponents(c),encoding(c),version(c)
    # owner(9s),date(26s),comment(80s),pad(3x),seek_pos(l),scratch(40s),
    # pad(4x)

    # note that between comment and seek_pos is a 3 byte pad
    # so that the long is @ a multiple of 4
    # also sparky always packs big-endian, hence >
    return struct.unpack('>10s 4c 9s 26s 80s 3x l 40s 4x', f.read(180))


def put_fileheader(f, fl):
    """
    Write fileheader list to file (180-bytes).
    """
    f.write(struct.pack('>10s 4c 9s 26s 80s 3x l 40s 4x', *fl))
    return


def fileheader2dic(header):
    """
    Convert a fileheader list into a Sparky parameter dictionary.
    """
    dic = dict()
    dic["ident"] = str(header[0].decode()).strip('\x00')
    dic["naxis"] = ord(header[1].decode())
    dic["ncomponents"] = ord(header[2].decode())
    dic["encoding"] = ord(header[3].decode())
    dic["version"] = ord(header[4].decode())
    dic["owner"] = str(header[5].decode()).strip('\x00')
    dic["date"] = str(header[6].decode()).strip('\x00')
    dic["comment"] = str(header[7].decode()).strip('\x00')
    dic["seek_pos"] = header[8]     # eof seek position
    dic["scratch"] = str(header[9].decode()).strip('\x00')
    return dic


def dic2fileheader(dic):
    """
    Convert a Sparky parameter dictionary into a fileheader list.
    """
    fl = [0] * 10
    fl[0] = dic["ident"].encode()
    fl[1] = chr(dic["naxis"]).encode()
    fl[2] = chr(dic["ncomponents"]).encode()
    fl[3] = chr(dic["encoding"]).encode()
    fl[4] = chr(dic["version"]).encode()
    fl[5] = dic["owner"].encode()
    fl[6] = dic["date"].encode()
    fl[7] = dic["comment"].encode()
    fl[8] = dic["seek_pos"]
    fl[9] = dic["scratch"].encode()
    return fl


# axisheader functions
def get_axisheader(f):
    """
    Get an axisheader from file and return a list.

    Only the first 44 bytes are examined, the NMR_PROCESSED and other header
    parameters are ignored since the current version of Sparky does not use
    them.

    """
    # axis header is described in ucsffile.cc
    # axis header is packed as follows
    # nucleus(6s),spectral_shift(h),npoints(I),size(I),bsize(I)
    # spectrometer_freq(f),spectral_width(f),xmtr_freq(f),zero_order(f),
    # first_order(f),first_pt_scale(f),ZEROS
    return struct.unpack('>6s h 3I 6f 84s', f.read(128))


def put_axisheader(f, al):
    """
    Write an axisheader list to file (128-bytes written).
    """
    f.write(struct.pack('>6s h 3I 6f 84s', *al))
    return


def axisheader2dic(header):
    """
    Convert an axisheader list into Sparky parameter axis dictionary.
    """
    dic = dict()
    dic["nucleus"] = str(header[0].decode()).strip('\x00')
    dic["spectral_shift"] = header[1]
    dic["npoints"] = header[2]
    dic["size"] = header[3]
    dic["bsize"] = header[4]
    dic["spectrometer_freq"] = header[5]
    dic["spectral_width"] = header[6]
    dic["xmtr_freq"] = header[7]
    dic["zero_order"] = header[8]
    dic["first_order"] = header[9]
    dic["first_pt_scale"] = header[10]
    dic["extended"] = header[11]
    return dic


def dic2axisheader(dic):
    """
    Convert a Sparky parameter axis diction into a axisherder list.
    """
    al = [0] * 12
    al[0] = dic["nucleus"].encode()
    al[1] = dic["spectral_shift"]
    al[2] = dic["npoints"]
    al[3] = dic["size"]
    al[4] = dic["bsize"]
    al[5] = dic["spectrometer_freq"]
    al[6] = dic["spectral_width"]
    al[7] = dic["xmtr_freq"]
    al[8] = dic["zero_order"]
    al[9] = dic["first_order"]
    al[10] = dic["first_pt_scale"]
    al[11] = dic["extended"]
    return al
